import warnings
import numpy
from simses.system.storage_system_dc import StorageSystemDC
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.system import SystemState
from simses.system.auxiliary.auxiliary import Auxiliary
from simses.system.thermal.ambient.ambient_thermal_model import AmbientThermalModel
from simses.system.thermal.ambient.location_temperature import LocationAmbientTemperature
from simses.system.thermal.ambient.user_battery_temperature import UserBatteryTemperatureProfile
from simses.system.thermal.model.system_thermal_model import SystemThermalModel
from simses.technology.storage import StorageTechnology


class NoSystemThermalModel(SystemThermalModel):
    """This model does nothing - keeps the system air temperature equal to ambient temperature"""

    LARGE_NUMBER = numpy.finfo(numpy.float64).max * 1e-100

    def __init__(self, ambient_thermal_model: AmbientThermalModel, general_config: GeneralSimulationConfig,
                 dc_systems: [StorageSystemDC]):
        super().__init__()
        self.__ambient_thermal_model = ambient_thermal_model
        if isinstance(self.__ambient_thermal_model, UserBatteryTemperatureProfile):
            self.__dc_systems: [StorageSystemDC] = dc_systems
            self.__storage_technologies: [StorageTechnology] = list()
            for dc_system in self.__dc_systems:  # Unpack storage technologies and DC/DC converters from StorageSystemDC
                self.__storage_technologies.append(dc_system.get_storage_technology())

        self.start_time = general_config.start
        if isinstance(ambient_thermal_model, LocationAmbientTemperature):
            warnings.warn('LocationAmbientTemperature is chosen with NoSystemThermalModel.\n'
                  'StorageTechnology temperatures will be initialized and will not be updated further.\n'
                  'Please Select ConstantAmbientTemperature if this is not desired.\n')
        self.__ambient_thermal_model = ambient_thermal_model
        self.__system_temperature = self.__ambient_thermal_model.get_initial_temperature()  # K
        self.__air_specific_heat = 1006  # J/kgK, cp (at constant pressure)
        # this is the internal air temperature within the container. Initialized with ambient temperature

    def calculate_temperature(self, time, state: SystemState, states: [SystemState]):
        self.__system_temperature = self.__ambient_thermal_model.get_temperature(time)
        if isinstance(self.__ambient_thermal_model, UserBatteryTemperatureProfile):
            # setting storage technology temperature for SIMSES simulation and plotting
            for storage_technology in self.__storage_technologies:
                storage_technology.state.temperature = self.__system_temperature

    def get_auxiliaries(self) -> [Auxiliary]:
        return list()

    def get_temperature(self):
        return self.__system_temperature

    def get_ambient_temperature(self) -> float:
        return self.__system_temperature

    def get_solar_irradiation_thermal_load(self) -> float:
        return 0.0

    def get_hvac_thermal_power(self) -> float:
        return 0.0

    def reset_profiles(self, ts_adapted: float) -> None:
        """
        Enables looping of the simulation beyond the original length of the time series for the AmbientThermalModel and
        SolarIrradiationModel
        """
        pass

    def update_air_parameters(self):
        pass

    def get_ol_temperature(self) -> float:
        pass

    def get_il_temperature(self) -> float:
        pass




from pandas import DataFrame, read_csv


class RoomToolReader:

    # __ROOM_TOOL_REPORT_FILE: str = 'report.csv'

    __MANUFACTURER: str = "Manufacturer"
    __MODEL: str = "Model"
    __FORMAT: str = "Format"
    __CONFIGURATION: str = "Configuration"
    __VOLTAGE_MAX: str = "Voltage_max (V)"
    __VOLTAGE_MIN: str = "Voltage_min (V)"
    __VOLTAGE_NOM_MODULE: str = "Nom. module voltage (V)"
    __WEIGHT: str = "Weight (kg)"
    __VOLUME: str = "Volume (m^3)"
    __COOLING_TYPE: str = "Cooling type"
    __ENERGY: str = "Energy (Wh)"

    def __init__(self, room_tool_file: str):
        # * We need a matching between selected cell and room tool
        # * we need to adapt not only serial / parallel but also voltage and capacity -> overwriting config
        # * How could we scale such a solution?
        # -> maybe we should not try to solve the interface on this low level but maybe on a high level e.g. batch
        #    processing
        self.__ROOM_TOOL_REPORT_FILE = room_tool_file
        self.__data_room_tool_report: DataFrame = read_csv(self.__ROOM_TOOL_REPORT_FILE, header=0)
        # self.__data: DataFrame = self.__read_room_tool_report(cell)

    def get_data_report(self):
        return self.__data_room_tool_report

    def __read_room_tool_report(self, cell: str):
        data: DataFrame = read_csv(self.__ROOM_TOOL_REPORT_FILE, header=0)

        # Select cell from dataset:
        cell_data: DataFrame = data.loc[(data[self.__MODEL] == cell[:-10])]


        # Workaround for selecting multiple reports from room tool
        # first_row: DataFrame = data.iloc[0]
        # print("RoomToolReader: Only first row is considered")
        # print("RoomToolReader: Cell matching is not available")
        # Note: Please consider a name matching of cell names between SimSES and room tool
        return cell_data

    def get_battery_scale(self) -> (int, int):
        # first_values_serial_parallel: str = self.__data[self.__CONFIGURATION]
        # serial: int = int(first_values_serial_parallel.partition("s")[0])
        # parallel: int = int(first_values_serial_parallel.partition("s")[2][0:-1])
        serial = int(self.__data["Cells in series"])
        parallel = int(self.__data["Cells in parallel"])

        return serial, parallel


    def get_energy(self) -> float:
        return float(self.__data[self.__ENERGY])

    def get_nominal_voltage(self) -> float:
        return float(self.__data[self.__VOLTAGE_NOM_MODULE])



from abc import ABC, abstractmethod


class CellFormat(ABC):

    """
    CellFormat inherits the dimensions of the cell and provides volume and surface area for each cell.
    """

    def __init__(self):
        self.__name: str = type(self).__name__

    @abstractmethod
    def get_volume(self) -> float:
        """

        Returns
        -------
        float:
            volume of one cell in m3
        """
        pass

    @abstractmethod
    def get_surface_area(self) -> float:
        """

        Returns
        -------
        float:
            surface area of one cell in m2
        """
        pass

    def get_name(self) -> str:
        """

        Returns
        -------
        str:
            name of the cell format
        """
        return self.__name



from configparser import ConfigParser

from simses.commons.config.data.data_config import DataConfig


class RedoxFlowDataConfig(DataConfig):

    def __init__(self, path: str = None, config: ConfigParser = None):
        super().__init__(path, config)
        self.__section: str = 'REDOX_FLOW_DATA'

    @property
    def redox_flow_data_dir(self) -> str:
        """Returns directory of redox flow data files"""
        return self.get_data_path(self.get_property(self.__section, 'REDOX_FLOW_DATA_DIR'))

    @property
    def rfb_rint_file_cell_stack(self) -> str:
        """Returns filename for internal resistance of the cell stack 5500W"""
        return self.redox_flow_data_dir + self.get_property(self.__section, 'CELL_STACK_RINT_FILE')

    @property
    def rfb_rint_file_hp_stack(self) -> str:
        """Returns filename for internal resistance of the high-performance stack 9500W"""
        return self.redox_flow_data_dir + self.get_property(self.__section, 'HIGH_PERFORMANCE_STACK_RINT_FILE')

    @property
    def redox_flow_hydrogen_evolution_dir(self) -> str:
        """Returns directory of redox flow hydrogen evolution current data files"""
        return self.get_data_path(self.get_property(self.__section, 'REDOX_FLOW_HYDROGEN_EVOLUTION_DATA'))

    @property
    def rfb_h2_evolution_schweiss_f1_file(self) -> str:
        """Returns filename for the hydrogen evolution of a RFB electrode F1 (source: Schweiss 2016)"""
        return self.redox_flow_hydrogen_evolution_dir + self.get_property(self.__section,
                                                                          'REDOX_FLOW_HYDROGEN_SCHWEISS_F1')

    @property
    def rfb_h2_evolution_schweiss_f2_file(self) -> str:
        """Returns filename for the hydrogen evolution of a RFB electrode F2 (source: Schweiss 2016)"""
        return self.redox_flow_hydrogen_evolution_dir + self.get_property(self.__section,
                                                                          'REDOX_FLOW_HYDROGEN_SCHWEISS_F2')

    @property
    def rfb_h2_evolution_schweiss_f3_file(self) -> str:
        """Returns filename for the hydrogen evolution of a RFB electrode F3 (source: Schweiss 2016)"""
        return self.redox_flow_hydrogen_evolution_dir + self.get_property(self.__section,
                                                                          'REDOX_FLOW_HYDROGEN_SCHWEISS_F3')

    @property
    def rfb_h2_evolution_schweiss_f4_file(self) -> str:
        """Returns filename for the hydrogen evolution of a RFB electrode F4 (source: Schweiss 2016)"""
        return self.redox_flow_hydrogen_evolution_dir + self.get_property(self.__section,
                                                                          'REDOX_FLOW_HYDROGEN_SCHWEISS_F4')

    @property
    def industrial_stack_1500w_shunt_current(self) -> str:
        """Returns filename for shunt current of industrial stack 1500W"""
        return self.redox_flow_data_dir + self.get_property(self.__section, 'INDUSTRIAL_STACK_1500W_SHUNT')

    @property
    def industrial_stack_9000w_shunt_current(self) -> str:
        """Returns filename for shunt current of industrial stack 9000W"""
        return self.redox_flow_data_dir + self.get_property(self.__section, 'INDUSTRIAL_STACK_9000W_SHUNT')

    @property
    def cell_stack_shunt_current(self) -> str:
        """Returns filename for shunt current of cell stack 5500W"""
        return self.redox_flow_data_dir + self.get_property(self.__section, 'CELL_STACK_SHUNT')

    @property
    def high_performance_stack_shunt(self) -> str:
        """Returns filename for shunt current of high-performance stack 9500W"""
        return self.redox_flow_data_dir + self.get_property(self.__section, 'HIGH_PERFORMANCE_STACK_SHUNT')

    @property
    def high_performance_stack_self_discharge(self) -> str:
        """Returns filename for self-discharge current of high-performance stack 9500W"""
        return self.redox_flow_data_dir + self.get_property(self.__section, 'HIGH_PERFORMANCE_STACK_SD_FILE')



from xml.etree import ElementTree
from xml.etree.ElementTree import Element

import numpy


class XmlReader:

    def __init__(self, filename: str):
        tree: ElementTree = ElementTree.parse(filename)
        self.__root: Element = tree.getroot()

    def show_all(self) -> None:
        self.show(self.__root)
        print('total amount of root items:', len(self.__root))
        self.show_subelement_names_of(self.__root)
        print('---')
        for elem in self.__root:
            elem: Element = elem
            self.show_all_subelements_of(elem)

    def show_subelement_names_of(self, element: Element) -> None:
        elements: [str] = self.get_subelement_names_of(element)
        print(elements)

    def get_subelement_names_of(self, element: Element) -> [str]:
        elements: [str] = list()
        for elem in element:
            elements.append(elem.tag)
        return elements

    def show_all_subelements_of(self, element: Element) -> None:
        if element:
            print('')
            self.show(element)
            for subelement in element:
                self.show_all_subelements_of(subelement)
        else:
            self.show(element)

    def show(self, element: Element):
        tag: str = element.tag
        attr: dict = element.attrib
        text: str = element.text
        print('tag:', tag, ', attr:', attr, ', text:', text)

    def get_element(self, name: str, element: Element = None) -> Element:
        if element is None:
            element = self.__root
        return element.find(name)

    def get_value_from(self, element: Element) -> str:
        if element:
            raise ValueError('XML element has children: ' + str(self.get_subelement_names_of(element)))
        return element.text

    def __parse_to_array(self, value: str, delimiter: str) -> numpy.ndarray:
        values: [float] = list()
        for val in value.split(delimiter):
            values.append(float(val))
        return numpy.array(values)

    def __parse_to_matrix(self, value: str, row_delimiter: str, column_delimiter: str) -> numpy.ndarray:
        matrix: [[float]] = list()
        for row in value.split(row_delimiter):
            values: [float] = list()
            for val in row.split(column_delimiter):
                values.append(float(val))
            matrix.append(values)
        return numpy.array(matrix)

    def parse(self, element: Element, row_delimiter: str = ';', column_delimiter: str = ',') -> numpy.ndarray:
        value: str = self.get_value_from(element)
        value = value.replace('\n', '')
        value = value.replace('\t', '')

        if row_delimiter in value:
            return self.__parse_to_matrix(value, row_delimiter, column_delimiter)
        else:
            return self.__parse_to_array(value, column_delimiter)

    # # find the first 'item' object
    # for elem in root:
    #     print(elem.find('item').get('name'))
    #
    # # find all "item" objects and print their "name" attribute
    # for elem in root:
    #     for subelem in elem.findall('item'):
    #         subelem: SubElement = subelem
    #         # if we don't need to know the name of the attribute(s), get the dict
    #         print(subelem.attrib)
    #
    #         # if we know the name of the attribute, access it directly
    #         print(subelem.get('name'))


if __name__ == '__main__':
    path: str = '../../data/lithium_ion/isea/i3Cell.xml'
    xml: XmlReader = XmlReader(path)
    custom_definitions: Element = xml.get_element('CustomDefinitions')
    xml.show_subelement_names_of(custom_definitions)
    print('---')
    ocv: Element = xml.get_element('MyOCV', custom_definitions)
    ocv_object: Element = xml.get_element('Object', ocv)
    ocv_data: Element = xml.get_element('LookupData', ocv_object)
    soc: Element = xml.get_element('MeasurementPointsRow', ocv_object)
    temperature: Element = xml.get_element('MeasurementPointsColumn', ocv_object)
    ocv_values: numpy.ndarray = xml.parse(ocv_data)
    soc_values: numpy.ndarray = xml.parse(soc)
    print(type(ocv_values), ocv_values)
    print(type(soc_values), soc_values)
    # xml.show_subelement_names_of(ocv_object)



import math

import pandas as pd
import scipy.interpolate

from simses.commons.config.data.battery import BatteryDataConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.log import Logger
from simses.commons.state.technology.lithium_ion import LithiumIonState
from simses.technology.lithium_ion.cell.electric.properties import ElectricalCellProperties
from simses.technology.lithium_ion.cell.format.abstract import CellFormat
from simses.technology.lithium_ion.cell.format.round_18650 import RoundCell18650
from simses.technology.lithium_ion.cell.thermal.properties import ThermalCellProperties
from simses.technology.lithium_ion.cell.type import CellType


class LGMJ1_NMC(CellType):
    """An LG_INR18650MJ1 is a special cell type and inherited by CellType"""

    __SOC_HEADER: str = 'SOC'
    __SOC_IDX: int = 0
    __TEMP_IDX: int = 1
    __C_RATE_IDX: int = 0
    __ETA_IDX: int = 1

    # Source: Datasheet LG INR18650 MJ1

    __CELL_VOLTAGE: float = 3.635  # V
    __CELL_CAPACITY: float = 3.5  # Ah
    __MAX_VOLTAGE: float = 4.2   # V
    __MIN_VOLTAGE: float = 2.5   # V

    __MIN_TEMPERATURE: float = -253.15  # K
    __MAX_TEMPERATURE: float = 333.15  # K
    __MAX_CHARGE_RATE: float = 1.0  # 1/h
    __MAX_DISCHARGE_RATE: float = 2.875  # 1/h

    __MASS: float = 0.049  # kg per cell

    # Source: Zilberman I., Sturm J., Jossen A.
    # Reversible self-discharge and calendar aging of 18650 nickel-rich silicon-graphite lithium-ion cells
    # Journal of Power Sources, https://doi.org/10.1016/j.jpowsour.2019.03.109

    __CELL_RESISTANCE: float = 0.04431  # Ohm # from lmo_daimler Source:
    __SELF_DISCHARGE_RATE: float = 0 # X.X%-soc per month in second step, e.g., 0.015 / (30.5 * 24 * 3600) for 1.5%
    # __SELF_DISCHARGE_RATE: float = 0.0027 / (30.5 * 24 * 3600)  # X.X%-soc per month in second step, e.g., 0.015 / (30.5 * 24 * 3600)

    # Source of specific heat:
    # Steinhardt M., Gillich E.I., Rheinfeld A., Kraft L., Spielbauer M., Bohlen O., Jossen A.
    # Low-effort determination of heat capacity and thermal conductivity for
    # cylindrical 18650 and 21700 lithium-ion cells
    # Journal of Energy Storage, https://doi.org/10.1016/j.est.2021.103065

    __SPECIFIC_HEAT: float = 999.45 # J/kgK
    __CONVECTION_COEFFICIENT: float = 15  # W/m2K convection coefficient of nmc sanyo ur18650e

    __COULOMB_EFFICIENCY: float = 1.0  # p.u.

    __ELECTRICAL_PROPS: ElectricalCellProperties = ElectricalCellProperties(__CELL_VOLTAGE, __CELL_CAPACITY,
                                                                            __MIN_VOLTAGE, __MAX_VOLTAGE,
                                                                            __MAX_CHARGE_RATE, __MAX_DISCHARGE_RATE,
                                                                            __SELF_DISCHARGE_RATE,
                                                                            __COULOMB_EFFICIENCY)
    __THERMAL_PROPS: ThermalCellProperties = ThermalCellProperties(__MIN_TEMPERATURE, __MAX_TEMPERATURE, __MASS,
                                                                   __SPECIFIC_HEAT, __CONVECTION_COEFFICIENT)
    __CELL_FORMAT: CellFormat = RoundCell18650()

    def __init__(self, voltage: float, capacity: float, soh: float, battery_config: BatteryConfig,
                 battery_data_config: BatteryDataConfig):
        super().__init__(voltage, capacity, soh, self.__ELECTRICAL_PROPS, self.__THERMAL_PROPS, self.__CELL_FORMAT,
                         battery_config)
        self.__log: Logger = Logger(type(self).__name__)

    # Source: J. Sturm, A. Rheinfeld, I. Zilberman, F.B. Spingler, S. Kosch, F. Frie, A. Jossen
    # Modeling and simulation of inhomogeneities in a 18650 nickel-rich, silicon-graphite lithium-ion cell
    # during fast charging
    # Journal of Power Sources, https://doi.org/10.1016/j.jpowsour.2018.11.043

    def get_open_circuit_voltage(self, battery_state: LithiumIonState) -> float:
        '''Parameters build with ocv fitting'''
        a1 = 15.2751
        a2 = 49.5206
        a3 = 1.63050
        a4 = 121.702
        b1 = -0.1848
        b2 = -0.6339
        k0 = 4.99612
        k1 = -4.9026
        k2 = 0.55138
        k3 = -1.9184
        k4 = -0.6234
        k5 = 0.15706

        soc = battery_state.soc

        ocv = k0 + \
              k1 / (1 + math.exp(a1 * (soc - b1))) + \
              k2 / (1 + math.exp(a2 * (soc - b2))) + \
              k3 / (1 + math.exp(a3 * (soc - 1))) +\
              k4 / (1 + math.exp(a4 * soc)) +\
              k5 * soc

        return ocv * self.get_serial_scale()

    def get_internal_resistance(self, battery_state: LithiumIonState) -> float:
        return float(self.__CELL_RESISTANCE / self.get_parallel_scale() * self.get_serial_scale())

    def close(self):
        self.__log.close()


import numpy as np
import pytest

from simses.commons.config.data.power_electronics import PowerElectronicsConfig
from simses.commons.utils.utilities import all_non_abstract_subclasses_of
from simses.system.power_electronics.acdc_converter.abstract_acdc_converter import AcDcConverter
from simses.system.power_electronics.acdc_converter.bonfiglioli import BonfiglioliAcDcConverter
from simses.system.power_electronics.acdc_converter.fix_efficiency import FixEfficiencyAcDcConverter
from simses.system.power_electronics.acdc_converter.no_loss import NoLossAcDcConverter
from simses.system.power_electronics.acdc_converter.notton import NottonAcDcConverter
from simses.system.power_electronics.acdc_converter.notton_loss import NottonLossAcDcConverter
from simses.system.power_electronics.acdc_converter.sinamics import Sinamics120AcDcConverter
from simses.system.power_electronics.acdc_converter.stable import M2bAcDcConverter
from simses.system.power_electronics.acdc_converter.stacked import AcDcConverterIdenticalStacked
from simses.system.power_electronics.acdc_converter.sungrow import SungrowAcDcConverter


class TestClassAcDc:

    max_power = 40000
    step = 10
    test_values = list(range(0, max_power, int(max_power / step)))
    abs_voltage = 20

    def make_converter(self, converter_subclass, max_power) -> AcDcConverter:
        if converter_subclass in [Sinamics120AcDcConverter]:
            return converter_subclass(max_power, PowerElectronicsConfig())
        elif converter_subclass in [AcDcConverterIdenticalStacked]:
            return AcDcConverterIdenticalStacked(3, 0.8, NottonAcDcConverter(max_power))
        elif converter_subclass in [BonfiglioliAcDcConverter, FixEfficiencyAcDcConverter,
                                    NoLossAcDcConverter, NottonAcDcConverter, NottonLossAcDcConverter, M2bAcDcConverter,
                                    SungrowAcDcConverter]:
            return converter_subclass(max_power)
        else:
            raise Exception('AcDcConverter is not tested')

    @pytest.fixture(scope="function")#module: gets only one instance
    def converter_subclass_list(self):
        return all_non_abstract_subclasses_of(AcDcConverter, [AcDcConverterIdenticalStacked])

    @pytest.mark.parametrize("ac_power", test_values)#sadly no function objects can be passed
    def test_sign_to_ac(self, ac_power, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
            assert uut.to_ac(ac_power, self.abs_voltage) == 0  # "charge"
            assert uut.to_ac(-ac_power, self.abs_voltage) <= -ac_power  # discharge

    @pytest.mark.parametrize("ac_power", test_values)
    def test_sign_to_dc(self, ac_power, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
            assert uut.to_dc(ac_power, self.abs_voltage) <= ac_power  # charge
            assert uut.to_dc(-ac_power, self.abs_voltage) == 0  # "discharge"

    @pytest.mark.parametrize("dc_power", test_values)
    def test_sign_to_ac_reverse(self, dc_power, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
            assert uut.to_ac_reverse(dc_power, self.abs_voltage) == 0  # "charge"
            assert uut.to_ac_reverse(-dc_power, self.abs_voltage) >= -dc_power  # discharge

    @pytest.mark.parametrize("dc_power", test_values)
    def test_sign_to_dc_reverse(self, dc_power, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
            assert uut.to_dc_reverse(dc_power, self.abs_voltage) >= dc_power  # charge
            assert uut.to_dc_reverse(-dc_power, self.abs_voltage) == 0  # "discharge"

    def test_max_power_to_ac(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                assert uut.to_ac(self.max_power * 2, self.abs_voltage) <= self.max_power
                assert uut.to_ac(-self.max_power * 2, self.abs_voltage) >= self.max_power

    def test_max_power_to_dc(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                assert uut.to_dc(self.max_power * 2, self.abs_voltage) <= self.max_power
                assert uut.to_dc(-self.max_power * 2, self.abs_voltage) >= self.max_power

    def test_max_power_to_ac_reverse(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                assert uut.to_ac_reverse(self.max_power * 2, self.abs_voltage) <= self.max_power
                assert uut.to_ac_reverse(-self.max_power * 2, self.abs_voltage) >= self.max_power

    def test_max_power_to_dc_reverse(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                assert uut.to_dc_reverse(self.max_power * 2, self.abs_voltage) <= self.max_power
                assert uut.to_dc_reverse(-self.max_power * 2, self.abs_voltage) >= self.max_power

    def test_to_ac_is_nan(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                print(converter_subclass, np.isnan(uut.to_ac(np.nan, self.abs_voltage)))
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                uut.to_ac(np.nan, self.abs_voltage)

    def test_to_dc_is_nan(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                uut.to_dc(np.nan, self.abs_voltage)

    def test_to_ac_reverse_is_nan(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                uut.to_ac_reverse(np.nan, self.abs_voltage)

    def test_to_dc_reverse_is_nan(self, converter_subclass_list):
        for converter_subclass in converter_subclass_list:
            with pytest.raises(Exception):
                uut: AcDcConverter = self.make_converter(converter_subclass, self.max_power)
                uut.to_dc_reverse(np.nan, self.abs_voltage)



from simses.commons.timeseries.average.average import Average
from simses.commons.timeseries.timevalue import TimeValue


class MeanAverage(Average):
    """
    MeanAverage provides a arithmetic mean behaviour of averaging.
    """

    def __init__(self):
        super().__init__()

    def average(self, data: [TimeValue]) -> float:
        values: [float] = list()
        for d in data:
            values.append(d.value)
        mean = sum(values) / len(values)
        return mean



import math

class Fan:
    """
    The fan class is an attribute of the HVAC classes
    Data sheet source:
    https://www.kaltra.com/wp-content/uploads/2020/01/SB_Delta_CW-CWU_Ver.5.0_EN.pdf
    self.__rated_airflow = 3.61  # m3/s
    self.__rated_power = 2560.0  # W
    https://catalog.pelonistechnologies.com/item/ac-axial-fans/ac-axial-fans-2/pn-23865
    self.__rated_airflow = 1.7855  # m3/s
    self.__rated_power = 980.0  # W

    """

    def __init__(self):

        # Fan system consists of multiple fan sub-units
        self.__number_units = 3
        self.__unit_rated_airflow = 1.7855  # m3/s
        self.__unit_rated_power = 980  # W

        self.__rated_airflow = self.__unit_rated_airflow * self.__number_units  # m3/s
        self.__rated_power = self.__unit_rated_power * self.__number_units  # W

        # Initialize
        self.__airflow = 0.0  # m3/s
        self.__electricity_consumption = 0.0  # W

    def run(self, airflow) -> None:
        # physical energy conversion equations here
        self.__airflow = airflow
        active_units = math.ceil(airflow / self.__unit_rated_airflow)
        self.__electricity_consumption = float(active_units * self.__unit_rated_power)

    @property
    def electricity_consumption(self) -> float:
        return self.__electricity_consumption

    @property
    def rated_airflow(self) -> float:
        return self.__rated_airflow

    @property
    def airflow(self) -> float:
        return self.__airfow



import os
import re
import webbrowser
from datetime import datetime

import pandas as pd
import plotly.graph_objects as go

from simses.analysis.evaluation.plotting.system_parameters import SystemParameterFigureGenerator
from simses.analysis.evaluation.abstract_evaluation import Evaluation
from simses.analysis.evaluation.plotting.plotly_plotting import PlotlyPlotting
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.result import EvaluationResult
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.state.parameters import SystemParameters


class EvaluationMerger:

    """
    EvaluationMerger merges all evaluations results and figures into one HTML file and opens it after finishing.
    """

    OUTPUT_NAME: str = 'Results.html'

    def __init__(self, result_path: str, config: GeneralAnalysisConfig, version: str):
        self.__file_name: str = os.path.join(result_path, self.OUTPUT_NAME)
        self.__simulation_name: str = result_path.split('/')[-3]
        self.__version: str = version
        simulation_time: str = result_path.split('/')[-2].split('_')[1].split('M')[0]
        self.__simulation_time: str = str(datetime.strptime(simulation_time, '%Y%m%dT%H%M%S'))
        self.__system_parameters: str = os.path.join(result_path,SystemParameters.get_file_name())
        self.__merge_results: bool = config.merge_analysis
        self.__logo_path: str = config.logo_file

    def merge(self, evaluations: [Evaluation]) -> None:
        """
        Writes html file from evaluation results and figures.

        Parameters:
            evaluations:   List of evaluations.
        """
        if not self.__merge_results:
            return
        with open(self.__file_name, 'w', encoding='utf-8') as outfile:
            outfile.write("<!DOCTYPE html><html><head></head><body>")
            outfile.write(self.__html_header())
            outfile.write(self.__html_style())
            self.__write_evaluations(evaluations, outfile)
            outfile.write("<br>")
            # outfile.write("<section><b>System parameters</b></section>")
            # outfile.write(Plotting.convert_to_html(SystemParameterFigureGenerator(self.__system_parameters).generate()))
            self.__write_system_parameters(outfile)
            outfile.write("</body></html>")
        webbrowser.open(self.__file_name, new=2)  # open in new tab

    def __write_evaluations(self, evaluations: [Evaluation], outfile) -> None:
        for evaluation in evaluations:
            evaluation: Evaluation = evaluation
            if evaluation.should_be_considered:
                section = re.sub('([A-Z0-9])', r' \1', evaluation.get_file_name())[:-4] # make filename human readable
                outfile.write("<section><b>" + section + "</b></section>")
                results = list()
                for result in evaluation.evaluation_results:
                    result: EvaluationResult = result
                    # outfile.write(result.to_console() + "<br>")
                    results.append(result.to_csv())
                results_df = pd.DataFrame(results, columns=EvaluationResult.get_header())
                outfile.write(Plotting.convert_to_html(self.__to_table(results_df)))
                for figure in evaluation.get_figures():
                    outfile.write(Plotting.convert_to_html(figure))
                outfile.write("<br><br>")

    def __write_system_parameters(self, outfile) -> None:
        with open(self.__system_parameters, 'r') as system_parameters:
            lines: [str] = system_parameters.readlines()
            for line in lines:
                outfile.write(line)

    def __to_table(self, result: pd.DataFrame) -> go.Table:
        """Returns EvaluationResult as a Table."""
        table_of_results = go.Figure(data=[go.Table(
            columnwidth=[400, 150, 150],
            header=dict(values=list(result.columns), fill_color='#0065BD', font_color="white", line_color="#0065BD"),
            cells=dict(values=[list(result.Description),list(result.Value),list(result.Unit)],
                       height=20, font_color="black", align='left', line_color=PlotlyPlotting.Color.SOC_BLUE,
                       fill_color='white'))])
        table_of_results.update_layout(width=700)
        return table_of_results

    def __html_header(self) -> str:
        header = '''<header>
        <p style="color:white;">Simulation name: ''' + self.__simulation_name + '''</p>
        <p style="color:white;">Date / Time:    ''' + self.__simulation_time + '''</p>
        <p style="color:white;">Version:    ''' + self.__version + '''</p>
        <img src=''' + self.__logo_path + ''' alt="SimSES" width=300>
        </header>'''

        return header

    def __html_style(self) -> str:
        style = '''<style>
            body{
                    background-color: white;
                    margin-left: 1cm;
                    margin-right: 1cm;

                    border: 5px solid #0065BD;
                    padding: 10px;
                    border-radius: 8px;
            }
            header {
                    background-color : #0065BD;
                    border: 1px solid white;
                    padding: 10px;
                    border-radius: 8px;
            }
            img {
                    display: block;
                    margin-left: auto;
                    margin-right: auto;
            }
            
       </style>'''

        return style



from configparser import ConfigParser
import pytest
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.system.thermal.ambient.location_temperature import LocationAmbientTemperature

config = None
general_config: GeneralSimulationConfig = GeneralSimulationConfig(config)
profile_config: ProfileConfig = ProfileConfig(config)

# Ensure if path from simulation.defaults.ini to ambient temperature model works
ambient_temperature_model_config: ConfigParser = ConfigParser()
ambient_temperature_model_config.add_section('STORAGE_SYSTEM')
ambient_temperature_model_config.set('STORAGE_SYSTEM', 'AMBIENT_TEMPERATURE_MODEL', 'LocationAmbientTemperature')
storage_system_config = StorageSystemConfig(ambient_temperature_model_config)
ambient_temperature_model = str(storage_system_config.ambient_temperature_model[StorageSystemConfig.AMBIENT_TEMPERATURE_TYPE])

start_time = 0
end_time = 100
sample_time = 1
time_step_range = range(start_time, end_time, sample_time)


@pytest.fixture(scope="function")
def uut() -> LocationAmbientTemperature:
    if ambient_temperature_model == LocationAmbientTemperature.__name__:
        return LocationAmbientTemperature(profile_config, general_config)
    else:
        raise Exception('Test failed: Check link between simulation config and factory.')


@pytest.mark.parametrize("time_step", time_step_range)
def test_get_temperature(time_step: float, uut: LocationAmbientTemperature):
    assert 329.85 > uut.get_temperature(time_step) > 190.35  # K,
    # not hotter/colder than the coldest temperature recorded on earth


def test_get_initial_temperature(uut: LocationAmbientTemperature):
    assert 329.85 > uut.get_initial_temperature() > 190.35  # K,
    # not hotter/colder than the coldest temperature recorded on earth



from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.cycle_detection.cycle_detector import CycleDetector
from simses.technology.lithium_ion.cell.type import CellType
from simses.technology.lithium_ion.degradation.calendar.lto_nmc import LTONMCCalendarDegradationModel
from simses.technology.lithium_ion.degradation.cyclic.lto_nmc import LTONMCCyclicDegradationModel
from simses.technology.lithium_ion.degradation.degradation_model import DegradationModel


class LTONMCDegradationModel(DegradationModel):

    def __init__(self, cell_type: CellType, cycle_detector: CycleDetector, battery_config: BatteryConfig):
        super().__init__(cell_type, LTONMCCyclicDegradationModel(cell_type, cycle_detector),
                         LTONMCCalendarDegradationModel(cell_type), cycle_detector, battery_config)



from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.commons.constants import Hydrogen
from simses.technology.hydrogen.control.thermal.thermal_controller import ThermalController


class IdealVarFlowThermalController(ThermalController):
    """
    This controller controls the temperature of the EL-stack by varing the input temperaure of the feeding water and
    in the second step adapt the mass flow of the water in order to reach the desired temperature
    """

    def __init__(self, config: ElectrolyzerConfig, heat_capacity_stack: float):
        super().__init__()
        self.__stack_temp_desire = config.desire_temperature  # °C
        self.__heat_capacity_stack = heat_capacity_stack  # J/K
        self.__delta_water_temperature = 5  # K
        self.__max_temperature_variation_rate = 1 / 60  #  K/s from: "Current status of water electrolysis for energy storage, grid balancing and sector coupling via power-to-gas and power-to-liquids: A review"
        self.__heat_desire = 0  # W
        self.__delta_water_temp_in = 0  # °C
        self.__h2o_flow = 0  # mol/s
        self.__current_dens_zero_counter = 0
        self.__shut_down_time = 0.5  # h
        self.__heat_control_on: bool = False
        self.__heat_h2o_calculated = 0

    def calculate(self, stack_temperature: float, heat_stack: float, timestep: float, current_dens: float) -> None:
        temp_diff = self.__stack_temp_desire - stack_temperature
        ideal_temp_variation_rate = temp_diff / timestep  # K/s
        if abs(ideal_temp_variation_rate) < abs(self.__max_temperature_variation_rate):
            temp_var_rate = ideal_temp_variation_rate
        else:
            temp_var_rate = ideal_temp_variation_rate / abs(ideal_temp_variation_rate) * self.__max_temperature_variation_rate
        self.__heat_desire = temp_var_rate * self.__heat_capacity_stack  # J/s
        self.__heat_control_on = self.check_control_status(current_dens, timestep)
        heat_water = self.__heat_desire - heat_stack
        if self.__heat_control_on:
            if heat_water > 0:  # stack needs to be heated
                self.__delta_water_temp_in = self.__delta_water_temperature  # K
                self.__h2o_flow = heat_water / (Hydrogen.HEAT_CAPACITY_WATER *
                                                Hydrogen.MOLAR_MASS_WATER * self.__delta_water_temp_in)  # mol/s
            elif heat_water < 0:  # stack needs to be cooled
                self.__delta_water_temp_in = - self.__delta_water_temperature  # K
                self.__h2o_flow = heat_water / (Hydrogen.HEAT_CAPACITY_WATER *
                                                Hydrogen.MOLAR_MASS_WATER * self.__delta_water_temp_in)  # mol/s
            else:
                self.__delta_water_temp_in = 0  # K
                self.__h2o_flow = 0  # mol/s
        else:
            self.__delta_water_temp_in = 0  # °K
            self.__h2o_flow = 0  # mol/s

    def check_control_status(self, current_dens, timestep) -> bool:
        if current_dens == 0:
            self.__current_dens_zero_counter += 1
        else:
            self.__current_dens_zero_counter = 0
        if self.__current_dens_zero_counter <= self.__shut_down_time * 3600 / timestep:
            return True
        else:
            return False

    def get_heat_control_on(self) -> bool:
        return self.__heat_control_on

    def get_delta_water_temp_in(self) -> float:
        return self.__delta_water_temp_in

    def get_h2o_flow(self) -> float:
        return self.__h2o_flow



from abc import ABC, abstractmethod

from simses.system.auxiliary.auxiliary import Auxiliary


class DcDcConverter(ABC):

    def __init__(self, intermediate_circuit_voltage: float):
        self.__intermediate_circuit_voltage: float = intermediate_circuit_voltage

    @abstractmethod
    def calculate_dc_current(self, dc_power: float, storage_voltage: float) -> None:
        """
        function to calculate the dc current

        Parameters
        ----------
        dc_power : dc input power in W
        storage_voltage : voltage of storage in V

        """
        pass

    @abstractmethod
    def reverse_calculate_dc_current(self, dc_power: float, storage_voltage: float) -> None:
        """
        function to calculate the dc current

        Parameters
        ----------
        dc_power :
            dc input power in W
        storage_voltage :
            voltage of storage in V

        """
        pass

    @property
    @abstractmethod
    def max_power(self) -> float:
        pass

    @property
    def intermediate_circuit_voltage(self) -> float:
        return self.__intermediate_circuit_voltage

    @property
    @abstractmethod
    def dc_power_loss(self) -> float:
        pass

    @property
    @abstractmethod
    def dc_power(self) -> float:
        pass

    @property
    @abstractmethod
    def dc_current(self) -> float:
        pass

    @property
    @abstractmethod
    def volume(self) -> float:
        """
        Volume of dc dc converter in m3

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def mass(self) -> float:
        """
        Mass of dc dc converter in kg

        Returns
        -------

        """
        pass

    @property
    @abstractmethod
    def surface_area(self) -> float:
        """
        Surface area of dc dc converter in m2

        Returns
        -------

        """
        pass

    @abstractmethod
    def get_auxiliaries(self) -> [Auxiliary]:
        pass

    @staticmethod
    def _is_charge(power: float) -> bool:
        return power > 0.0

    @abstractmethod
    def close(self) -> None:
        """
        Closing all open resources in dcdc converter

        Parameters
        ----------

        Returns
        -------

        """

        pass



from abc import ABC, abstractmethod


class PowerProfile(ABC):
    """
    Power Profile for the energy management system of the ESS.
    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def next(self, time: float) -> float:
        """
        provides the power for the next step of a specific load or generator (e.g. Household load profile)

        Parameters
        ----------
        time : current timestamp of the simulation

        Returns
        -------
        float
            power for the next step
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """
        Closing all open resources in power profile

        Parameters
        ----------

        Returns
        -------

        """
        pass



import numpy as np
from simses.analysis.data.lithium_ion import LithiumIonData
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotly_plotting import PlotlyPlotting
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.analysis.evaluation.technical.technical_evaluation import TechnicalEvaluation
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.config.simulation.battery import BatteryConfig
from simses.commons.state.technology.lithium_ion import LithiumIonState

class LithiumIonTechnicalEvaluation(TechnicalEvaluation):

    __title_overview = 'SOC, internal resistance, temperature and capacity'
    __title_current = 'Current and voltage'
    __title_degradation = 'Degradation of capacity and increase of internal resistance'

    def __init__(self, data: LithiumIonData, config: GeneralAnalysisConfig, battery_config: BatteryConfig, path: str):
        super().__init__(data, config)
        title_extension: str = ' for lithium-ion system ' + self.get_data().id
        self.__title_overview += title_extension
        self.__title_current += title_extension
        self.__title_degradation += title_extension
        self.__result_path = path

    def evaluate(self):
        super().evaluate()
        self.append_result(EvaluationResult(Description.Technical.EQUIVALENT_FULL_CYCLES, Unit.NONE, self.equivalent_full_cycles))
        self.append_result(EvaluationResult(Description.Technical.DEPTH_OF_DISCHARGE, Unit.PERCENTAGE, self.depth_of_discharges))
        self.append_result(EvaluationResult(Description.Technical.STORAGE_TECHNOLOGY_TEMPORAL_UTILIZATION, Unit.NONE, self.storage_technology_temporal_utilization))
        self.append_result((EvaluationResult(Description.Technical.C_RATE_CHARGING, Unit.ONE_PER_HOUR, self.c_rate_mean_charge)))
        self.append_result((EvaluationResult(Description.Technical.C_RATE_DISCHARGING, Unit.ONE_PER_HOUR, self.c_rate_mean_discharge)))

        # self.append_result(EvaluationResult(Description.Technical.ENERGY_THROUGHPUT, Unit.KWH, self.energy_throughput))
        # self.append_time_series(LithiumIonState.VOLTAGE, self.get_data().voltage)
        self.print_results()

    @property
    def c_rate_mean_charge(self) -> float:

        """
        Calculates the average c-rate while charging

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            mean c-rate when charging
        """
        data: LithiumIonData = self.get_data()
        charge_current_mean = np.mean(data.current[data.current > 0])
        battery_energy = 1000*data.initial_capacity # in Wh

        nominal_voltage = float(data.nominal_voltage)
        battery_capacity_ah = battery_energy/nominal_voltage
        c_rate_charging = charge_current_mean / battery_capacity_ah

        charge_power_mean = np.mean(data.power[data.power > 0])
        e_rate_charging = charge_power_mean/battery_energy # based on power and energy instead of current and capacity

        return c_rate_charging

    @property
    def c_rate_mean_discharge(self) -> float:

        """
        Calculates the average c-rate while discharging

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            mean c-rate when discharging
        """
        data: LithiumIonData = self.get_data()
        discharge_current_mean = np.mean(data.current[data.current < 0])
        battery_energy = 1000*data.initial_capacity # in Wh

        nominal_voltage = data.nominal_voltage
        battery_capacity_ah = battery_energy / nominal_voltage
        c_rate_discharging = -discharge_current_mean / battery_capacity_ah

        discharge_power_mean = np.mean(data.power[data.power < 0])
        e_rate_discharging = -discharge_power_mean/battery_energy # based on power and energy instead of current and capacity

        return c_rate_discharging


    def plot(self) -> None:
        self.__plot_overview()
        self.__plot_current_voltage()
        self.__plot_degradation()

    def __plot_overview(self) -> None:
        data: LithiumIonData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.__title_overview, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=LithiumIonState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.soc, label=LithiumIonState.SOC, color=PlotlyPlotting.Color.SOC_BLUE,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.capacity * 1000, label=LithiumIonState.CAPACITY,
                          color=PlotlyPlotting.Color.SOH_GREEN, linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.resistance, label=LithiumIonState.INTERNAL_RESISTANCE,
                          color=PlotlyPlotting.Color.RESISTANCE_BLACK, linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.temperature, label=LithiumIonState.TEMPERATURE,
                          color=PlotlyPlotting.Color.TEMPERATURE_RED, linestyle=PlotlyPlotting.Linestyle.SOLID))
        plot.subplots(xaxis=xaxis, yaxis=yaxis)
        self.extend_figures(plot.get_figures())

    def __plot_current_voltage(self) -> None:
        data: LithiumIonData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.__title_current, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=LithiumIonState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.current, label=LithiumIonState.CURRENT,
                          color=PlotlyPlotting.Color.RESISTANCE_BLACK, linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.voltage, label=LithiumIonState.VOLTAGE,
                          color=PlotlyPlotting.Color.TEMPERATURE_RED, linestyle=PlotlyPlotting.Linestyle.SOLID))
        plot.lines(xaxis=xaxis, yaxis=yaxis, secondary=[1])
        self.extend_figures(plot.get_figures())

    def __plot_degradation(self) -> None:
        data: LithiumIonData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.__title_degradation, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=LithiumIonState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.capacity_loss_calendar, label=LithiumIonState.CAPACITY_LOSS_CALENDRIC,
                          color=PlotlyPlotting.Color.DC_POWER_GREEN,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.capacity_loss_cyclic, label=LithiumIonState.CAPACITY_LOSS_CYCLIC,
                          color=PlotlyPlotting.Color.AC_POWER_BLUE,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.resistance_increase_calendar, label=LithiumIonState.RESISTANCE_INCREASE_CALENDRIC,
                          color=PlotlyPlotting.Color.RED,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.resistance_increase_cyclic, label=LithiumIonState.RESISTANCE_INCREASE_CYCLIC,
                          color=PlotlyPlotting.Color.GREEN,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        # plot.lines(xaxis=xaxis, yaxis=yaxis, secondary=[2, 3])
        plot.bar(yaxis=yaxis, bars=int(len(yaxis) / 2))
        self.extend_figures(plot.get_figures())

    # @property
    # def capacity_remaining(self) -> float:
    #     data: LithiumIonData = self.get_data()
    #     return data.state_of_health[-1] * 100



import pandas
from simses.analysis.data.abstract_data import Data
from simses.analysis.utils import get_positive_values_from, get_negative_values_from
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.state.system import SystemState


class SystemData(Data):

    """
    Provides time series data from SystemState
    """

    def __init__(self, config: GeneralSimulationConfig, data: pandas.DataFrame):
        super().__init__(config, data)

    @property
    def id(self) -> str:
        return str(int(self._get_first_value(SystemState.SYSTEM_AC_ID))) + '.' + \
               str(int(self._get_first_value(SystemState.SYSTEM_DC_ID)))

    @property
    def time(self):
        return self._get_data(SystemState.TIME)

    @property
    def power(self):
        return self._get_data(SystemState.AC_POWER_DELIVERED)

    @property
    def aux_power(self):
        return self._get_data(SystemState.AUX_LOSSES)

    @property
    def solar_thermal_load(self):
        return self._get_data(SystemState.SOLAR_THERMAL_LOAD)

    @property
    def hvac_thermal_power(self):
        return self._get_data(SystemState.HVAC_THERMAL_POWER)

    @property
    def pe_losses(self):
        return self._get_data(SystemState.PE_LOSSES)

    @property
    def total_pe_losses_energy(self):
        return sum(self.pe_losses * self.convert_watt_to_kWh)

    @property
    def total_aux_losses_energy(self):
        return sum(self.aux_power * self.convert_watt_to_kWh)

    @property
    def ac_pe_power(self):
        return self.power - self.aux_power

    @property
    def ac_pe_power_charging_energy_series(self):
        return get_positive_values_from(self.power - self.aux_power) * self.convert_watt_to_kWh

    @property
    def ac_pe_power_discharging_energy_series(self):
        return get_negative_values_from(self.power - self.aux_power - self.pe_losses) * self.convert_watt_to_kWh

    @property
    def ac_pe_charging_energy(self):
        return sum(self.ac_pe_power_charging_energy_series)

    @property
    def ac_pe_discharging_energy(self):
        return sum(self.ac_pe_power_discharging_energy_series)

    @property
    def aux_energy_charging(self):
        return self.charge_energy - self.ac_pe_charging_energy

    @property
    def energy_difference(self):
        soc = self._get_difference(SystemState.SOC)
        capacity = self.initial_capacity
        return soc * capacity

    @property
    def soc(self):
        return self._get_data(SystemState.SOC)

    @property
    def capacity(self):
        return self._get_data(SystemState.CAPACITY) / 1000.0

    @property
    def state_of_health(self):
        return self._get_data(SystemState.SOH)

    @property
    def temperature(self):
        return self._get_data(SystemState.TEMPERATURE)

    @property
    def ol_temperature(self):
        return self._get_data(SystemState.OL_TEMPERATURE)

    @property
    def il_temperature(self):
        return self._get_data(SystemState.IL_TEMPERATURE)

    @property
    def ambient_temperature(self):
        return self._get_data(SystemState.AMBIENT_TEMPERATURE)

    @property
    def storage_fulfillment(self):
        return self._get_data(SystemState.AC_FULFILLMENT)

    @property
    def dc_power(self):
        return self._get_data(SystemState.DC_POWER_INTERMEDIATE_CIRCUIT)

    @property
    def initial_energy_content(self):
        return self._get_first_value(SystemState.SOC) * self.initial_capacity

    @property
    def final_energy_content(self):
        return self._get_last_value(SystemState.CAPACITY) * self._get_last_value(SystemState.SOC) / 1000

    @property
    def dc_power_charging_energy_series(self):
        return get_positive_values_from(self.dc_power) * self.convert_watt_to_kWh

    @property
    def dc_power_charging_energy(self):
        return sum(self.dc_power_charging_energy_series)

    @property
    def dc_power_discharging_energy_series(self):
        return get_negative_values_from(self.dc_power) * self.convert_watt_to_kWh

    @property
    def dc_power_discharging_energy(self):
        return sum(self.dc_power_discharging_energy_series)

    @property
    def dc_voltage(self):
        return self._get_data(SystemState.DC_VOLTAGE_CIRCUIT)

    @property
    def dc_power_storage(self):
        return self._get_data(SystemState.DC_POWER_STORAGE)

    @property
    def dc_power_storage_charging_energy_series(self):
        return get_positive_values_from(self.dc_power_storage) * self.convert_watt_to_kWh

    @property
    def dc_power_storage_charging_energy(self):
        return sum(self.dc_power_storage_charging_energy_series)

    @property
    def dc_power_storage_discharging_energy_series(self):
        return get_negative_values_from(self.dc_power_storage) * self.convert_watt_to_kWh

    @property
    def dc_power_storage_discharging_energy(self):
        return sum(self.dc_power_storage_discharging_energy_series)

    @property
    def dc_power_additional(self):
        return self._get_data(SystemState.DC_POWER_ADDITIONAL)

    @property
    def dc_power_loss(self):
        return self._get_data(SystemState.DC_POWER_LOSS)

    @property
    def storage_technology_loss_power(self):
        return self._get_data(SystemState.STORAGE_POWER_LOSS)

    @property
    def storage_technology_loss_energy_series(self):
        return self.storage_technology_loss_power * self.convert_watt_to_kWh

    @property
    def storage_technology_loss_energy(self):
        return sum(self.storage_technology_loss_energy_series)

    @property
    def ac_power_target(self):
        return self._get_data(SystemState.AC_POWER)

    @property
    def is_top_level_system(self) -> bool:
        return self.id == '0.0'

    @classmethod
    def get_system_data(cls, path: str, config: GeneralSimulationConfig) -> list:
        system_data: [pandas.DataFrame] = cls._get_system_data_for(path, SystemState, SystemState.TIME,
                                                                   SystemState.SYSTEM_AC_ID, SystemState.SYSTEM_DC_ID)
        res: [SystemData] = list()
        for data in system_data:
            res.append(SystemData(config, data))
        # solution for duplicate information if only one ac system was simulated
        if len(res) == 2:
            systems = res[:]
            for system in systems:
                if not system.is_top_level_system:
                    res.remove(system)
        Data.sort_by_id(res)
        return res



from simses.commons.config.simulation.electrolyzer import ElectrolyzerConfig
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.control.pressure.no_pressure_controller import NoPressureController
from simses.technology.hydrogen.electrolyzer.pressure.abstract_model import PressureModel


class AlkalinePressureModel(PressureModel):

    def __init__(self, config: ElectrolyzerConfig):
        super().__init__()
        self.__pressure_cathode_el = config.desire_pressure_cathode  # barg
        self.__pressure_anode_el = config.desire_pressure_anode  # barg
        if self.__pressure_cathode_el != self.__pressure_anode_el:
            raise Exception('Current alkaline electrolzyer assumes same pressure on cathode and anode. '
                            'Please check your config.')
        self.__pressure_cathode_desire_el = config.desire_pressure_cathode  # barg
        self.__pressure_anode_desire_el = config.desire_pressure_anode  # barg
        self.__hydrogen_outflow = 0  # mol/s
        self.__oxygen_outflow = 0  # mol/s
        self.__pressure_controller = NoPressureController()

    def calculate(self, time, state: ElectrolyzerState) -> None:
        # Express anode/cathode pressure relative to atmospheric pressure, for implementation with existing hydrogen storage model
        # self.__pressure_anode_el: float = state.pressure_anode - 1     # in barg
        # self.__pressure_cathode_el: float = state.pressure_cathode - 1   # in barg
        hydrogen_produced = state.hydrogen_production
        oxygen_produced = state.oxygen_production
        self.__hydrogen_outflow = self.__pressure_controller.calculate_n_h2_out(self.__pressure_cathode_el, self.__pressure_cathode_desire_el, hydrogen_produced, 0)
        self.__oxygen_outflow = self.__pressure_controller.calculate_n_o2_out(self.__pressure_anode_el, self.__pressure_anode_desire_el, oxygen_produced)

    def get_pressure_anode(self) -> float:
        return self.__pressure_anode_el

    def get_pressure_cathode(self) -> float:
        return self.__pressure_cathode_el

    def get_h2_outflow(self) -> float:
        return self.__hydrogen_outflow

    def get_o2_outflow(self) -> float:
        return self.__oxygen_outflow

    def get_h2o_outflow_cathode(self) -> float:
        return 0

    def get_h2o_outflow_anode(self) -> float:
        return 0



from configparser import ConfigParser

from simses.commons.config.data.fuel_cell import FuelCellDataConfig
from simses.commons.config.simulation.fuel_cell import FuelCellConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.constants import Hydrogen
from simses.commons.log import Logger
from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.technology.hydrogen.fuel_cell.pressure.no_pressure import NoPressureModel
from simses.technology.hydrogen.fuel_cell.pressure.pressure_model import PressureModel
from simses.technology.hydrogen.fuel_cell.stack.jupiter import JupiterFuelCell
from simses.technology.hydrogen.fuel_cell.stack.no_fuel_cell import NoFuelCell
from simses.technology.hydrogen.fuel_cell.stack.pem import PemFuelCell
from simses.technology.hydrogen.fuel_cell.stack.stack_model import \
    FuelCellStackModel
from simses.technology.hydrogen.fuel_cell.thermal.no_thermal_model import NoThermalModel
from simses.technology.hydrogen.fuel_cell.thermal.simple import SimpleThermalModel
from simses.technology.hydrogen.fuel_cell.thermal.thermal_model import ThermalModel


class FuelCellFactory:

    def __init__(self, config: ConfigParser):
        self.__log: Logger = Logger(type(self).__name__)
        self.__config_general: GeneralSimulationConfig = GeneralSimulationConfig(config)
        self.__config_fuel_cell: FuelCellConfig = FuelCellConfig(config)
        self.__config_fuel_cell_data: FuelCellDataConfig = FuelCellDataConfig()

    def create_state(self, system_id: int, storage_id: int, temperature: float,
                     fuel_cell: FuelCellStackModel) -> FuelCellState:
        state: FuelCellState = FuelCellState(system_id, storage_id)
        state.time = self.__config_general.start
        # state.power = 0  # W
        state.temperature = temperature - 273.15  # K -> °C
        state.power_loss = 0  # W
        state.fulfillment = 1.0  # p.u.
        state.soh = 1.0  # p.u.
        state.convection_heat = 0  # W
        state.hydrogen_use = 0  # mols
        state.hydrogen_inflow = 0  # mol/s
        state.oxygen_use = 0  # mol/s
        state.oxygen_inflow = 0  # mol/s
        state.pressure_cathode = Hydrogen.EPS  # barg
        state.pressure_anode = Hydrogen.EPS  # barg
        state.voltage = 1  # stays at 1 so that electrolyzer and fuel cell always see the power indepentently from the voltage a timestep before
        fuel_cell.calculate(0)
        state.voltage = fuel_cell.get_voltage()
        state.current = fuel_cell.get_current()
        state.current_density = fuel_cell.get_current_density()
        # state.power = fuel_cell.get_power()
        return state

    def create_fuel_cell_stack(self, fuel_cell: str, fuel_cell_maximal_power: float) -> FuelCellStackModel:
        if fuel_cell == PemFuelCell.__name__:
            self.__log.debug('Creating fuel cell as ' + fuel_cell)
            return PemFuelCell(fuel_cell_maximal_power, self.__config_fuel_cell_data)
        elif fuel_cell == JupiterFuelCell.__name__:
            self.__log.debug('Creating fuel cell as ' + fuel_cell)
            return JupiterFuelCell(fuel_cell_maximal_power, self.__config_fuel_cell_data)
        elif fuel_cell == NoFuelCell.__name__:
            self.__log.debug('Creating fuel cell as ' + fuel_cell)
            return NoFuelCell()
        else:
            options: [str] = list()
            options.append(PemFuelCell.__name__)
            options.append(JupiterFuelCell.__name__)
            options.append(NoFuelCell.__name__)
            raise Exception('Specified fuel cell ' + fuel_cell + ' is unknown. '
                                                                 'Following options are available: ' + str(options))

    def create_thermal_model(self, fuel_cell: FuelCellStackModel, temperature: float) -> ThermalModel:
        # TODO For which fuel cell do you want to include which thermal model?
        if type(fuel_cell).__name__ in [JupiterFuelCell.__name__]:
            return SimpleThermalModel(temperature)
        else:
            return NoThermalModel()

    def create_pressure_model(self, fuel_cell: FuelCellStackModel) -> PressureModel:
        # TODO For which fuel cell do you want to include which pressure model?
        return NoPressureModel()
        # if pressure_model_fc == NoPressureModelFc.__name__:
        #     self.__log.debug('Creatubg electrolyzer pressure model as ' + pressure_model_fc)
        #     return NoPressureModelFc()
        # else:
        #     options: [str] = list()
        #     options.append(NoPressureModelFc.__name__)
        #     raise Exception('Specified thermal model ' + pressure_model_fc + 'is unknown. '
        #                                                                      'Following options are available: ' + str(
        #         options))

    def close(self):
        self.__log.close()



import os
from configparser import ConfigParser
from datetime import datetime, timezone
import pytest
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.utils.utilities import remove_file
from simses.system.housing.abstract_housing import Housing
from simses.system.housing.forty_ft_container import FortyFtContainer
from simses.system.housing.twenty_ft_container import TwentyFtContainer
from simses.system.thermal.solar_irradiation.location import \
    LocationSolarIrradiationModel

DELIMITER: str = ','
HEADER: str = '# Unit: W/m^2\n # Latitude: 54\n # Longitude: 15'
START: int = 0
END: int = 60
STEP: int = 1


def create_file(value) -> str:
    filename: str = os.path.join(os.path.dirname(__file__), 'mockup_ghi_profile.csv')  # ghi - global horizontal irradiance
    with open(filename, mode='w') as file:
        file.write(HEADER + '\n')
        # add bad inputs
        file.write('0;0')
        file.write('\n')
        file.write('#\n')
        file.write('"""\n')
        file.write('"\n')
        file.write('   #\n')
        # add good inputs
        tstmp = START
        while tstmp <= END:
            file.write(str(tstmp) + DELIMITER + str(value) + '\n')
            tstmp += STEP
    return filename


def create_general_config() -> GeneralSimulationConfig:
    conf: ConfigParser = ConfigParser()
    conf.add_section(GeneralSimulationConfig.SECTION)
    date = datetime.fromtimestamp(START, timezone.utc)
    conf.set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.START, date.strftime("%Y-%m-%d %H:%M:%S"))
    date = datetime.fromtimestamp(END, timezone.utc)
    conf.set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.END, date.strftime("%Y-%m-%d %H:%M:%S"))
    conf.set(GeneralSimulationConfig.SECTION, GeneralSimulationConfig.TIME_STEP, str(STEP))
    return GeneralSimulationConfig(conf)


def create_data_config(filename: str) -> ProfileConfig:
    conf: ConfigParser = ConfigParser()
    conf.add_section(ProfileConfig.SECTION)
    conf.set(ProfileConfig.SECTION, ProfileConfig.THERMAL_PROFILE_DIR, os.path.dirname(filename))
    conf.set(ProfileConfig.SECTION, ProfileConfig.GLOBAL_HORIZONTAL_IRRADIATION_PROFILE, os.path.basename(filename))
    return ProfileConfig(conf)


@pytest.fixture(scope='function')
def uut(global_horizontal_irradiance, housing):
    filename = create_file(global_horizontal_irradiance)
    if housing == TwentyFtContainer.__name__:
        housing_model: Housing = TwentyFtContainer(None)
    elif housing == FortyFtContainer.__name__:
        housing_model: Housing = FortyFtContainer(None)
    uut = LocationSolarIrradiationModel(create_data_config(filename), create_general_config(), housing_model)
    yield uut
    uut.close()
    remove_file(filename)


@pytest.mark.parametrize('global_horizontal_irradiance, time_factor, housing, result',
                         [
                             (100, 1, TwentyFtContainer.__name__, 100),
                             (200, 1, TwentyFtContainer.__name__, 200),
                             (300, 1, FortyFtContainer.__name__, 300),
                             (400, 1, FortyFtContainer.__name__, 400),
                             (500, 2, FortyFtContainer.__name__, 500),
                             (600, 5, TwentyFtContainer.__name__, 600),
                             (0, 5, TwentyFtContainer.__name__, 0),
                             (10, 4, TwentyFtContainer.__name__, 10),
                             (20, 10, TwentyFtContainer.__name__, 20)
                         ]
                         )
def test_methods(time_factor, housing, result, uut):
    tstmp = START
    if housing == TwentyFtContainer.__name__:
        housing_model: Housing = TwentyFtContainer(None)
    elif housing == FortyFtContainer.__name__:
        housing_model: Housing = FortyFtContainer(None)
    while tstmp < END:
        assert uut.get_heat_load(tstmp) <= 1300 * housing_model.outer_layer.surface_area_total * housing_model.outer_layer.absorptivity
        assert abs(uut.get_global_horizontal_irradiance(tstmp) - result) <= 1e-10
        tstmp += STEP * time_factor







from abc import ABC, abstractmethod

from simses.commons.log import Logger
from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.degradation.calendar.calendar_degradation import \
    CalendarDegradationModel
from simses.technology.hydrogen.electrolyzer.degradation.cyclic.cyclic_degradation import \
    CyclicDegradationModel


class DegradationModel(ABC):
    """
    Model for the degradation_model_el behavior of the electrolyzer by analysing the resistance increase and exchange current density.
    """

    def __init__(self, cyclic_degradation_model: CyclicDegradationModel, calendar_degradation_model: CalendarDegradationModel):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)
        self.__calendar_degradation_model: CalendarDegradationModel = calendar_degradation_model
        self.__cyclic_degradation_model: CyclicDegradationModel = cyclic_degradation_model

    def update(self, time:float, state: ElectrolyzerState) -> None:
        """
        Updating the resistance and exchange current density of the electrolyzer through the degradation_model_el model.

        Parameters
        ----------
        time : float
            Current timestamp.
        state : HydrogenState
            Current state of the hydrogen storage system.

        Returns
        -------

        """
        self.calculate_degradation(time, state)

        # Resistance increase
        if state.resistance_increase_cyclic + self.__cyclic_degradation_model.get_resistance_increase() <= 0:  # check that cyclic resistance is not getting smaller than its inital value
            state.resistance_increase_cyclic = 0
        else:
            state.resistance_increase_cyclic += self.__cyclic_degradation_model.get_resistance_increase()
        state.resistance_increase_calendar = self.__calendar_degradation_model.get_resistance_increase()
        state.resistance_increase = state.resistance_increase_cyclic + state.resistance_increase_calendar

        self.__log.debug('Resistance increase cyclic: ' + str(state.resistance_increase_cyclic))
        self.__log.debug('Resistance increase calendric: ' + str(state.resistance_increase_calendar))
        self.__log.debug('Resistance increase total: ' + str(state.resistance_increase))

        # Exchangecurrent decrease
        state.exchange_current_decrease_cyclic = self.__cyclic_degradation_model.get_exchange_current_dens_decrease()
        state.exchange_current_decrease_calendar = self.__calendar_degradation_model.get_exchange_current_dens_decrease()
        state.exchange_current_decrease = state.exchange_current_decrease_cyclic + \
                                          state.exchange_current_decrease_calendar
        self.__log.debug('Exchange current density decrease cyclic: ' + str(state.exchange_current_decrease_cyclic))
        self.__log.debug('Exchange current density decrease calendric: ' + str(state.exchange_current_decrease_calendar))
        self.__log.debug('Exchange current density decrease total: ' + str(state.exchange_current_decrease))

        self.calculate_soh(state)
        state.soh = self.get_soh_el()

    def calculate_degradation(self, time: float, state: ElectrolyzerState) -> None:
        """
        Calculates degradation parameters of the specific electrolyzer

        Parameters
        ----------
        time : float
            Current timestamp.
        state : HydrogenState
            Current state of the Electrolyzer.

        Returns
        -------
        """

        self.__calendar_degradation_model.calculate_degradation(state)
        self.__cyclic_degradation_model.calculate_degradation(time, state)

    @abstractmethod
    def calculate_soh(self, state: ElectrolyzerState) -> None:
        """
        Calculates the SOH of the electrolyzer

        Parameters
        ----------
        state : HydrogenState
            Current state of health of the electrolyzer.

        Returns
        -------

        """
    pass

    @abstractmethod
    def get_soh_el(self):
        pass


    def close(self) -> None:
        """
        Closing all resources in degradation_model_el model

        Returns
        -------

        """
        self.__log.close()
        self.__calendar_degradation_model.close()
        self.__cyclic_degradation_model.close()



import pandas as pd

from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig


class ParametersAlkalineElectrolyzerFit:

    def __init__(self, electrolyzer_data_config: ElectrolyzerDataConfig):
        file = electrolyzer_data_config.alkaline_electrolyzer_fit_para_file
        correction_parameters = pd.read_csv(file)
        self.__corr_parameters = correction_parameters.iloc[:, 0]

    # Regression Parameters for Zirfon Membrane Resistance
    @property
    def zirfon1(self) -> float:
        return 2.085e-5

    @property
    def zirfon2(self) -> float:
        return -0.01603

    @property
    def zirfon3(self) -> float:
        return 3.188

    # Curve Fitting Correction Factors
    @property
    def ac(self) -> float:
        return self.__corr_parameters[0]

    @property
    def bc(self) -> float:
        return self.__corr_parameters[1]

    @property
    def cc(self) -> float:
        return self.__corr_parameters[2]

    @property
    def aa(self) -> float:
        return self.__corr_parameters[3]

    @property
    def ba(self) -> float:
        return self.__corr_parameters[4]

    @property
    def ca(self) -> float:
        return self.__corr_parameters[5]

    @property
    def r1(self) -> float:
        return self.__corr_parameters[6]

    @property
    def r2(self) -> float:
        return self.__corr_parameters[7]

    @property
    def r3(self) -> float:
        return self.__corr_parameters[8]

    @property
    def a1(self) -> float:
        return self.__corr_parameters[9]

    @property
    def a2(self) -> float:
        return self.__corr_parameters[10]



from simses.commons.log import Logger
from simses.commons.timeseries.interpolation.interpolation import Interpolation
from simses.commons.timeseries.timevalue import TimeValue


class LastValue(Interpolation):

    """
    Returns last value of given time range
    """

    def __init__(self):
        super().__init__()
        self.__log: Logger = Logger(type(self).__name__)

    def interpolate(self, time: float, recent: TimeValue, last: TimeValue) -> float:
        if time >= recent.time:
            return recent.value
        elif time >= last.time:
            return last.value
        else:
            self.__log.error('No value found for ' + str(time))
            return last.value

    def close(self) -> None:
        self.__log.close()



from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.technology.hydrogen.electrolyzer.stack.stack_model import ElectrolyzerStackModel


class NoElectrolyzer(ElectrolyzerStackModel):

    def __init__(self):
        super(NoElectrolyzer, self).__init__()

    def calculate(self, power: float, state: ElectrolyzerState) -> None:
        pass

    def get_reference_voltage_eol(self, resistance_increase: float, exchange_currentdensity_decrease: float) -> float:
        return 0

    def get_current(self):
        return 0

    def get_current_density(self):
        return 0

    def get_voltage(self):
        return 2

    def get_hydrogen_production(self):
        return 0

    def get_oxygen_production(self):
        return 0

    def get_water_use(self):
        return 0

    def get_number_cells(self):
        return 0

    def get_geom_area_stack(self):
        return 0

    def get_nominal_stack_power(self):
        return 0

    def get_heat_capacity_stack(self):
        return 0

    def get_partial_pressure_h2(self):
        return 0

    def get_partial_pressure_o2(self):
        return 0

    def get_sat_pressure_h2o(self):
        return 0

    def get_water_in_stack(self):
        return 0

    def get_nominal_current_density(self):
        return 0

    def get_thermal_resistance_stack(self):
        return 0

    def close(self):
        pass



import numpy as np
from simses.analysis.data.system import SystemData
from simses.analysis.evaluation.plotting.axis import Axis
from simses.analysis.evaluation.plotting.plotly_plotting import PlotlyPlotting
from simses.analysis.evaluation.plotting.plotter import Plotting
from simses.analysis.evaluation.plotting.sankey_diagram import SankeyDiagram
from simses.analysis.evaluation.plotting.sunburst_diagram import SunburstDiagram
from simses.analysis.evaluation.result import EvaluationResult, Description, Unit
from simses.analysis.evaluation.technical.technical_evaluation import TechnicalEvaluation
from simses.analysis.utils import get_positive_values_from, get_sum_for, get_negative_values_from, get_min_for, \
    get_max_for
from simses.commons.config.analysis.general import GeneralAnalysisConfig
from simses.commons.log import Logger
from simses.commons.state.system import SystemState


class SystemTechnicalEvaluation(TechnicalEvaluation):

    __power_title = 'System power'
    __soc_title = 'System SOC'
    __system_thermal_title = 'System thermal parameters'
    __additional_title = 'Additional DC power'
    __power_loss_title = 'Power losses'
    __voltage_title = 'Intermediate circuit voltage'
    __energy_flows_title = 'Energy flows'
    __loss_distribution_title = 'Losses distribution'

    def __init__(self, data: SystemData, config: GeneralAnalysisConfig, path: str):
        super().__init__(data, config)
        self.__log: Logger = Logger(type(self).__name__)
        title_extension: str = ' for system ' + self.get_data().id
        self.__power_title += title_extension
        self.__soc_title += title_extension
        self.__power_loss_title += title_extension
        self.__system_thermal_title += title_extension
        self.__energy_flows_title += title_extension
        self.__additional_title += title_extension
        self.__loss_distribution_title += title_extension
        self.__result_path = path
        self._set_numpy_err_handling()

    def _set_numpy_err_handling(self):
        """Log numpy's `divide by zero` warnings instead of printing"""
        def err_handler(type, flag):
            self.__log.warn("Numpy floating point error (%s), with flag %s" % (type, flag))
        np.seterrcall(err_handler)
        np.seterr(all='call')

        # self.__battery_config = storage_system_config
        # self.__test: float = sum([float(storage_system_ac[StorageSystemConfig.AC_SYSTEM_POWER])
        #                         for storage_system_ac in storage_system_config.storage_systems_ac]) * 1e-3
        # self.__system_initial_capacity: float = sum([float(storage_system_config.storage_technologies[storage_system_dc
        #                 [StorageSystemConfig.DC_SYSTEM_STORAGE]][StorageSystemConfig.STORAGE_CAPACITY])
        #                    for storage_system_dc in storage_system_config.storage_systems_dc]) * 1e-3
        # self.__system_initial_capacity: float = self.__get_initial_capacity_from(storage_system_config)

    # def __get_initial_capacity_from(self, storage_system_config: StorageSystemConfig) -> float:
    #     capacity: float = 0.0
    #     storage_technologies: dict = storage_system_config.storage_technologies
    #     for storage_system_dc in storage_system_config.storage_systems_dc:
    #         storage_name: str = storage_system_dc[StorageSystemConfig.DC_SYSTEM_STORAGE]
    #         if storage_name in storage_technologies.keys():
    #             capacity += float(storage_technologies[storage_name][StorageSystemConfig.STORAGE_CAPACITY])
    #     return capacity * 1e-3

    def evaluate(self):
        super().evaluate()
        data: SystemData = self.get_data()
        total_dcdc_charge_efficiency: float    = self.total_dcdc_efficiency_charge()
        total_dcdc_discharge_efficiency: float = self.total_dcdc_efficiency_discharge()
        total_dcdc_efficiency = total_dcdc_charge_efficiency * total_dcdc_discharge_efficiency / 100.0
        total_acdc_charge_efficiency: float    = self.total_acdc_efficiency_charge()
        total_acdc_discharge_efficiency: float = self.total_acdc_efficiency_discharge()
        total_acdc_efficiency = total_acdc_charge_efficiency * total_acdc_discharge_efficiency / 100.0
        total_pe_efficiency = total_dcdc_efficiency * total_acdc_efficiency / 100.0

        self.append_result(EvaluationResult(Description.Technical.E_RATE_CHARGING, Unit.ONE_PER_HOUR, self.e_rate_mean_charge()))
        self.append_result(EvaluationResult(Description.Technical.E_RATE_DISCHARGING, Unit.ONE_PER_HOUR, self.e_rate_mean_discharge()))
        self.append_result(EvaluationResult(Description.Technical.DCDC_EFFICIENCY_CHARGE, Unit.PERCENTAGE, total_dcdc_charge_efficiency))
        self.append_result(EvaluationResult(Description.Technical.DCDC_EFFICIENCY_DISCHARGE, Unit.PERCENTAGE, total_dcdc_discharge_efficiency))
        self.append_result(EvaluationResult(Description.Technical.DCDC_EFFICIENCY, Unit.PERCENTAGE, total_dcdc_efficiency))
        self.append_result(EvaluationResult(Description.Technical.ACDC_EFFICIENCY_CHARGE, Unit.PERCENTAGE, total_acdc_charge_efficiency))
        self.append_result(EvaluationResult(Description.Technical.ACDC_EFFICIENCY_DISCHARGE, Unit.PERCENTAGE, total_acdc_discharge_efficiency))
        self.append_result(EvaluationResult(Description.Technical.ACDC_EFFICIENCY, Unit.PERCENTAGE, total_acdc_efficiency))
        self.append_result(EvaluationResult(Description.Technical.PE_EFFICIENCY, Unit.PERCENTAGE, total_pe_efficiency))
        self.append_result(EvaluationResult(Description.Technical.MAX_LOAD_DC_POWER_ADDITIONAL, Unit.WATT, self.max_load_dc_power_additional))
        self.append_result(EvaluationResult(Description.Technical.MAX_GENERATION_DC_POWER_ADDITIONAL, Unit.WATT, self.max_generation_dc_power_additional))
        self.append_result(EvaluationResult(Description.Technical.SYSTEM_TEMPORAL_UTILIZATION, Unit.NONE, self.system_temporal_utilization()))
        # self.append_time_series(SystemState.DC_POWER_ADDITIONAL, data.dc_power_additional)
        # self.append_time_series(SystemState.DC_POWER_STORAGE, data.dc_power_storage)
        # self.append_time_series(SystemState.SOC, data.soc)
        self.print_results()

    def plot(self) -> None:
        self.__power_plotting()
        self.__soc_plotting()
        self.__power_loss_plotting()
        self.__system_thermal_plotting()
        # self.__voltage_plotting()
        self.__additional_dc_power_plotting()
        self.__energy_flows_plotting()
        self.__loss_distribution_plotting()

    def __soc_plotting(self):
        data: SystemData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.__soc_title, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=SystemState.TIME)
        yaxis: [Axis] = [Axis(data.soc, label=SystemState.SOC, color=PlotlyPlotting.Color.SOC_BLUE)]
        plot.lines(xaxis, yaxis)
        plot.histogram(xaxis=xaxis, yaxis=yaxis)
        self.extend_figures(plot.get_figures())

    def __voltage_plotting(self):
        data: SystemData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.__voltage_title, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=SystemState.TIME)
        yaxis: [Axis] = [Axis(data.dc_voltage, label=SystemState.DC_VOLTAGE_CIRCUIT, color=PlotlyPlotting.Color.SOC_BLUE)]
        plot.lines(xaxis, yaxis)
        plot.histogram(xaxis=xaxis, yaxis=yaxis)
        self.extend_figures(plot.get_figures())

    def __power_loss_plotting(self):
        data: SystemData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.__power_loss_title, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=SystemState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.pe_losses, label=SystemState.PE_LOSSES, color=PlotlyPlotting.Color.RED,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.dc_power_loss, label=SystemState.DC_POWER_LOSS, color=PlotlyPlotting.Color.HEAT_ORANGE,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.storage_technology_loss_power, label=SystemState.STORAGE_POWER_LOSS, color=PlotlyPlotting.Color.BLACK,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.aux_power, label=SystemState.AUX_LOSSES,
                          color=PlotlyPlotting.Color.BLUE, linestyle=PlotlyPlotting.Linestyle.SOLID))
        plot.lines(xaxis=xaxis, yaxis=yaxis)
        self.extend_figures(plot.get_figures())

    def __additional_dc_power_plotting(self):
        data: SystemData = self.get_data()
        if sum(abs(data.dc_power_additional)) > 0.0:
            plot: Plotting = PlotlyPlotting(title=self.__additional_title, path=self.__result_path)
            xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=SystemState.TIME)
            yaxis: [Axis] = [Axis(data.dc_power_additional, label=SystemState.DC_POWER_ADDITIONAL)]
            plot.lines(xaxis, yaxis)
            self.extend_figures(plot.get_figures())

    def __power_plotting(self):
        data: SystemData = self.get_data()
        plot: Plotting = PlotlyPlotting(title=self.__power_title, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=SystemState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data.dc_power_storage, label=SystemState.DC_POWER_STORAGE,
                          color=PlotlyPlotting.Color.DC_POWER_GREEN,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data.power, label=SystemState.AC_POWER_DELIVERED,
                          color=PlotlyPlotting.Color.AC_POWER_BLUE,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data.dc_power, label=SystemState.DC_POWER_INTERMEDIATE_CIRCUIT,
                          color=PlotlyPlotting.Color.RED,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data.ac_power_target, label=SystemState.AC_POWER,
                          color=PlotlyPlotting.Color.GREEN,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        plot.lines(xaxis=xaxis, yaxis=yaxis)
        plot.histogram(xaxis=xaxis, yaxis=yaxis)
        self.extend_figures(plot.get_figures())

    def __system_thermal_plotting(self):

        data: SystemData = self.get_data()
        if abs(sum(data.hvac_thermal_power)) > 1:  # HVAC power only present in scenarios with housing and solar irradiation
            plot: Plotting = PlotlyPlotting(title=self.__system_thermal_title, path=self.__result_path)
            xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=SystemState.TIME)
            yaxis: [Axis] = list()

            yaxis.append(Axis(data=data.solar_thermal_load, label=SystemState.SOLAR_THERMAL_LOAD,
                              color=PlotlyPlotting.Color.POWER_YELLOW, linestyle=PlotlyPlotting.Linestyle.SOLID))
            yaxis.append(Axis(data=-1*data.hvac_thermal_power, label=SystemState.HVAC_THERMAL_POWER, color=PlotlyPlotting.Color.BRIGHT_BLUE,
                              linestyle=PlotlyPlotting.Linestyle.SOLID))
            plot.subplots(xaxis=xaxis, yaxis=yaxis)
            self.extend_figures(plot.get_figures())

        plot3: Plotting = PlotlyPlotting(title=self.__system_thermal_title, path=self.__result_path)
        xaxis: Axis = Axis(data=Plotting.format_time(data.time), label=SystemState.TIME)
        yaxis: [Axis] = list()
        yaxis.append(Axis(data=data.ambient_temperature, label=SystemState.AMBIENT_TEMPERATURE,
                          color=PlotlyPlotting.Color.BRIGHT_GREEN, linestyle=PlotlyPlotting.Linestyle.SOLID))
        if sum(data.ol_temperature) > 1:
            yaxis.append(Axis(data=data.ol_temperature, label=SystemState.OL_TEMPERATURE, color=PlotlyPlotting.Color.YELLOW,
                              linestyle=PlotlyPlotting.Linestyle.SOLID))
            yaxis.append(Axis(data=data.il_temperature, label=SystemState.IL_TEMPERATURE, color=PlotlyPlotting.Color.GREY,
                              linestyle=PlotlyPlotting.Linestyle.SOLID))
        yaxis.append(Axis(data=data.temperature, label=SystemState.TEMPERATURE, color=PlotlyPlotting.Color.TEMPERATURE_RED,
                          linestyle=PlotlyPlotting.Linestyle.SOLID))
        plot3.lines(xaxis=xaxis, yaxis=yaxis)
        self.extend_figures(plot3.get_figures())

    def __energy_flows_plotting(self):
        data: SystemData = self.get_data()
        sankey_diagram: SankeyDiagram = SankeyDiagram(data, title=self.__energy_flows_title, path=self.__result_path)
        node_links: dict = sankey_diagram.create_node_links()
        plot: Plotting = PlotlyPlotting(self.__energy_flows_title, self.__result_path)
        plot.sankey_diagram(node_links)
        self.extend_figures(plot.get_figures())

    def __loss_distribution_plotting(self):
        data: SystemData = self.get_data()
        sunburst_diagram: SunburstDiagram = SunburstDiagram(data, title=self.__loss_distribution_title, path=self.__result_path)
        diagram_parameters: dict = sunburst_diagram.create_parameters()
        plot: Plotting = PlotlyPlotting(self.__loss_distribution_title, self.__result_path)
        plot.sunburst_diagram(diagram_parameters)
        self.extend_figures(plot.get_figures())

    @property
    def max_load_dc_power_additional(self) -> float:
        data: SystemData = self.get_data()
        dc_power = min(0.0, get_min_for(data.dc_power_additional))
        return abs(dc_power)

    @property
    def max_generation_dc_power_additional(self) -> float:
        data: SystemData = self.get_data()
        dc_power = max(0.0, get_max_for(data.dc_power_additional))
        return abs(dc_power)

    # @property
    # def capacity_remaining(self) -> float:
    #     data: SystemData = self.get_data()
    #     return data.state_of_health[-1] * 100.0


    def __total_charge_efficiency(self, power_in: np.ndarray, power_out: np.ndarray) -> float:
        """
        Calculates the total charge efficiency.

        Parameters
        ----------
        power_in: np.ndarray
            power coming into the converter
        power_out: np.ndarray
            power coming out of the converter

        Returns
        -------
        total_charge_efficiency: float
        """
        total_power_in = power_in[power_in > 0].sum()
        total_power_out = power_out[power_out > 0].sum()
        return 100.0 * (total_power_out / total_power_in)

    def __total_discharge_efficiency(self, power_in: np.ndarray, power_out: np.ndarray) -> float:
        """
        Calculates the total discharge efficiency

        Parameters
        ----------
        power_in: np.ndarray
            power coming into the converter
        power_out: np.ndarray
            power coming out of the converter

        Returns
        -------
        total_discharge_efficiency: float
        """
        total_power_in = power_in[power_in < 0].sum()
        total_power_out = power_out[power_out < 0].sum()
        return 100.0 * (total_power_out / total_power_in)

    def total_dcdc_efficiency_charge(self) -> float:
        data: SystemData = self.get_data()
        dc_power_intermediate_circuit = data.dc_power + data.dc_power_additional
        return self.__total_charge_efficiency(dc_power_intermediate_circuit, data.dc_power_storage)

    def total_dcdc_efficiency_discharge(self) -> float:
        data: SystemData = self.get_data()
        dc_power_intermediate_circuit = data.dc_power + data.dc_power_additional
        return self.__total_discharge_efficiency(data.dc_power_storage, dc_power_intermediate_circuit)

    def total_acdc_efficiency_charge(self) -> float:
        data: SystemData = self.get_data()
        return self.__total_charge_efficiency(data.ac_pe_power, data.dc_power)

    def total_acdc_efficiency_discharge(self) -> float:
        data: SystemData = self.get_data()
        return self.__total_discharge_efficiency(data.dc_power, data.ac_pe_power)

    def system_temporal_utilization(self) -> float:
        """
        calculates and returns system temporal utilization in p.u.
        :return: system temporal utilization as float
        """
        data: SystemData = self.get_data()
        ac_power: np.ndarray = data.power
        t_simulation = len(ac_power)
        t_operation = len(np.where(ac_power != 0)[0])
        return t_operation/t_simulation

    @staticmethod
    def __contains_only_nan(data: np.ndarray) -> bool:
        """
        If data array contains only NaN, method returns True; False otherwise.

        Parameters
        ----------
        data :

        Returns
        -------

        """
        return len(data) == 0 or np.isnan(data).all()

    def e_rate_mean_charge(self) -> float:
        """
        Calculates the average e-rate while charging
        E-rate is calculated by avg. power (W) divided by energy (Wh)

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            mean e-rate when charging
        """

        data: SystemData = self.get_data()
        battery_energy = 1000 * data.initial_capacity  # in Wh
        charge_power_mean = np.mean(data.power[data.power > 0])
        e_rate_charging = charge_power_mean / battery_energy  # based on power and energy instead of current and capacity

        return e_rate_charging

    def e_rate_mean_discharge(self) -> float:
        """
        Calculates the average e-rate while discharging.
        E-rate is calculated by avg. power (W) divided by energy (Wh)

        Parameters
        ----------
            data : simulation results

        Returns
        -------
        float:
            mean e-rate when discharging
        """

        data: SystemData = self.get_data()
        battery_energy = 1000 * data.initial_capacity  # in Wh
        discharge_power_mean = np.mean(data.power[data.power < 0])
        e_rate_discharging = -discharge_power_mean / battery_energy  # based on power and energy instead of current and capacity

        return e_rate_discharging

    def close(self) -> None:
        self.__log.close()
        super().close()



from abc import ABC, abstractmethod

from simses.commons.state.technology.fuel_cell import FuelCellState


class ThermalModel(ABC):

    def __init__(self):
        super().__init__()

    def update(self, time: float, state: FuelCellState) -> None:
        """Updating temperature of electrolyzer stack in hydrogen state"""
        self.calculate(time, state)
        state.temperature = self.get_temperature()
        state.water_flow_fc = self.get_water_flow_stack()

    @abstractmethod
    def calculate(self, time: float, state: FuelCellState) -> None:
        pass

    @abstractmethod
    def get_temperature(self) -> float:
        pass

    @abstractmethod
    def get_water_flow_stack(self) -> float:
        pass

    @abstractmethod
    def get_power_water_heating(self) -> float:
        pass

    @abstractmethod
    def get_pump_power(self) -> float:
        pass

    def close(self) -> None:
        """Closing all resources in thermal model"""
        pass



from simses.commons.config.data.electrolyzer import ElectrolyzerDataConfig
from simses.technology.hydrogen.electrolyzer.stack.pem_analytic.electrolyzer_model import \
    PemElectrolyzerMultiDimAnalytic


class PemElectrolyzerMultiDimAnalyticPtlCoating(PemElectrolyzerMultiDimAnalytic):
    """ PEM Electrolyzer with Pt-coated Ti-PTL at anode site, which prevents the cyclic increase of its resistance """
    def __init__(self, electrolyzer_maximal_power: float, electrolyzer_data_config: ElectrolyzerDataConfig):
        super().__init__(electrolyzer_maximal_power, electrolyzer_data_config)



from configparser import ConfigParser
import pytest
from simses.commons.config.simulation.system import StorageSystemConfig
from simses.system.auxiliary.heating_ventilation_air_conditioning.fix_cop_hvac import \
    FixCOPHeatingVentilationAirConditioning
from random import randrange


class TestFixCOPHeatingVentilationAirConditioning:

    # Ensure if path from simulation.defaults.ini to ambient temperature model works
    hvac_model_config: ConfigParser = ConfigParser()
    hvac_model_config.add_section('STORAGE_SYSTEM')
    hvac_model_config.set('STORAGE_SYSTEM', 'HVAC', 'constant_hvac,FixCOPHeatingVentilationAirConditioning,5000,25')
    storage_system_config = StorageSystemConfig(hvac_model_config)
    hvac = 'constant_hvac'
    hvac_model = storage_system_config.hvac[hvac][StorageSystemConfig.HVAC_TYPE]
    hvac_configuration = storage_system_config.hvac[hvac]

    set_point_deviation = 20  # °C
    air_mass = 100  # in kg
    air_specific_heat = 1006  # J/kgK
    temperature = 295  # K
    temperature_timestep = 60

    air_pressure = 1 * 10 ** 5  # Pa, change if container is pressurized
    universal_gas_constant = 8.314  # J/kgK
    molecular_weight_air = 28.965 * 10 ** -3  # kg/mol
    individual_gas_constant = universal_gas_constant / molecular_weight_air  # J/kgK
    air_density = air_pressure / (individual_gas_constant * temperature)

    max_thermal_power: int = int(storage_system_config.hvac[hvac][StorageSystemConfig.HVAC_POWER])  # in W
    set_point_temperature: int = int(storage_system_config.hvac[hvac][StorageSystemConfig.HVAC_TEMPERATURE_SETPOINT])  # in °C

    internal_air_temps = list(range(set_point_temperature - set_point_deviation, set_point_temperature + set_point_deviation, 1))  # K
    ambient_temps = list(range(260, 320, 1))  # K

    test_values = list()
    for x in internal_air_temps:
        for y in ambient_temps:
            test_values.append((x + 273.15, y))

    def create_model(self):
        if self.hvac_model == FixCOPHeatingVentilationAirConditioning.__name__:
            return FixCOPHeatingVentilationAirConditioning(self.hvac_configuration)

    @pytest.mark.parametrize('internal_air_temperature, ambient_air_temp', test_values)
    def test_run_air_conditioning(self, internal_air_temperature, ambient_air_temp):
        uut: FixCOPHeatingVentilationAirConditioning = self.create_model()
        uut.update_air_parameters(self.air_mass, self.air_specific_heat, self.air_density)  # Dummy value for mass of air

        uut.run_air_conditioning([internal_air_temperature], self.temperature_timestep, ambient_air_temp)

        thermal_power = uut.get_thermal_power()
        tms_temperature_dead_band = uut.get_temperature_dead_band()

        set_point_temperature = self.set_point_temperature + 273.15  # K

        if internal_air_temperature > set_point_temperature + tms_temperature_dead_band:
            # Cooling
            assert thermal_power >= 0
            if internal_air_temperature < ambient_air_temp:
                # Fresh air cooling not possible
                assert abs(thermal_power) <= self.max_thermal_power
        elif internal_air_temperature < set_point_temperature - tms_temperature_dead_band:
            # Heating
            assert thermal_power < 0
            if internal_air_temperature > ambient_air_temp:
                # Fresh air heating not possible
                assert abs(thermal_power) <= self.max_thermal_power
        else:
            assert thermal_power == 0



import pandas as pd
import scipy.interpolate

from simses.commons.config.data.redox_flow import RedoxFlowDataConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.degradation.capacity_degradation import CapacityDegradationModel
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class VariableHydrogenCurrent(CapacityDegradationModel):
    """Model that considers the reduction in capacity due to a hydrogen current for a redox flow battery. The hydrogen
     current is dependent on the state-of-charge."""

    def __init__(self, capacity: float, stack_module: StackModule, redox_flow_data_config: RedoxFlowDataConfig):
        super().__init__(capacity)
        self.__log: Logger = Logger(type(self).__name__)
        self.__stack_module = stack_module
        """
        The hydrogen evolution current density used in this model is based on data from Schweiss et al. for single cell 
        measurements. The values overestimate the expected capacity degradation due to hydrogen evolution. The model is 
        still in need of a published hydrogen current density of an industrial-sized stack. 
        
        In the paper of Schweiss et al. different carbon felt electrodes regarding precursor and graphitization level 
        are investigated. The hydrogen current density is dependent on the state of charge.
        F1: Carbon precursor is Rayon, traces of iron and nickel
        F2: Carbon precursor is Rayon, heat treatment above 2000 °C (Ar) and thermal surface oxidation (air, 750 °C) for
            better wettability
        F3: Carbon precursor is PAN, traces of iron and nickel
        F4: Carbon precursor is PAN, heat treatment above 2000 °C (Ar) and thermal surface oxidation (air, 750 °C) for
            better wettability
        Order of hydrogen evolution: F4 < F2 < F1 < F3
        
        Literature source: 
        Schweiss, Ruediger, Alexander Pritzl, and Christian Meiser. "Parasitic hydrogen evolution at different carbon 
        fiber electrodes in vanadium redox flow batteries." Journal of the Electrochemical Society 163.9 (2016): A2089. 
        """
        self.__HYDROGEN_EVOLUTION_CURRENT_FILE = redox_flow_data_config.rfb_h2_evolution_schweiss_f4_file
        self.__hydrogen_current_mat = pd.read_csv(self.__HYDROGEN_EVOLUTION_CURRENT_FILE)
        self.__soc_arr = self.__hydrogen_current_mat.iloc[:, 0]
        self.__i_h2 = self.__hydrogen_current_mat.iloc[:, 1]

    def get_capacity_degradation(self, time: float, redox_flow_state: RedoxFlowState):
        time_passed = time - redox_flow_state.time
        hydrogen_current = (self.__get_hydrogen_current_density(redox_flow_state) *
                            self.__stack_module.get_specific_cell_area())
        self.__log.debug('The hydrogen generation current (per cell) is: ' + str(hydrogen_current))
        capacity_degradation = (hydrogen_current * time_passed * self.__stack_module.get_nominal_voltage_cell() / 3600 *
                                self.__stack_module.get_cell_per_stack() * self.__stack_module.get_serial_scale() *
                                self.__stack_module.get_parallel_scale())
        return capacity_degradation  # Wh

    def __get_hydrogen_current_density(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Determines the hydrogen evolution current from table values.

        Parameters
        ----------
        redox_flow_state : RedoxFlowState
            Current state of redox_flow.

        Returns
        -------
        float :
            Hydrogen evolution current density in A/cm^2
        """
        i_h2_interp1d = scipy.interpolate.interp1d(self.__soc_arr, self.__i_h2, kind='linear')
        hydrogen_current_density = i_h2_interp1d(redox_flow_state.soc)  # A/cm^2
        self.__log.debug('The hydrogen generation current density is: ' + str(hydrogen_current_density))
        return hydrogen_current_density

    def close(self):
        super().close()
        self.__log.close()
        self.__stack_module.close()



from abc import ABC, abstractmethod

from simses.commons.state.technology.electrolyzer import ElectrolyzerState
from simses.commons.state.technology.fuel_cell import FuelCellState
from simses.system.auxiliary.compression.hydrogen_isentrop import \
    HydrogenIsentropCompressor


class HydrogenStorage(ABC):

    __TEMPERATURE: float = 40  # C

    def __init__(self):
        super().__init__()
        self.__compressor = HydrogenIsentropCompressor()

    def update_from(self, time_difference: float, electrolyzer_state: ElectrolyzerState, fuel_cell_state: FuelCellState) -> None:
        """

        Parameters
        ----------
        time :
        electrolyzer_state :
        fuel_cell_state :

        Returns
        -------

        """
        self.calculate_soc(time_difference, electrolyzer_state.hydrogen_production - fuel_cell_state.hydrogen_use)
        self.__compressor.calculate_compression_power(electrolyzer_state.hydrogen_outflow, electrolyzer_state.pressure_anode + 1,
                                                      self.get_tank_pressure() + 1, self.__TEMPERATURE)
        electrolyzer_state.power_compressor = self.__compressor.get_compression_power()

    @abstractmethod
    def calculate_soc(self, time_diff: float, hydrogen_net_flow: float) -> None:
        pass

    @abstractmethod
    def get_soc(self) -> float:
        pass

    @abstractmethod
    def get_capacity(self) -> float:
        pass

    @abstractmethod
    def get_tank_pressure(self) -> float:
        pass

    def get_auxiliaries(self):
        return [self.__compressor]

    @abstractmethod
    def close(self) -> None:
        pass



import numpy
from simses.system.housing.abstract_housing import Housing


class NoHousing(Housing):
    LARGE_NUMBER = numpy.finfo(numpy.float64).max * 1e-100

    def __init__(self):
        super().__init__()

    @property
    def internal_volume(self) -> float:
        return self.LARGE_NUMBER

    @property
    def azimuth(self) -> float:
        return 0.0

    @property
    def albedo(self) -> float:
        return 0.0




import datetime
import datetime as dt
import time
import numpy as np
from pytz import timezone
from simses.commons.config.simulation.profile import ProfileConfig
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.profile.file import FileProfile
from simses.system.housing.abstract_housing import Housing
from simses.system.housing.twenty_ft_container import TwentyFtContainer
from simses.system.thermal.ambient.location_temperature import LocationAmbientTemperature
from simses.system.thermal.solar_irradiation.solar_irradiation_model import SolarIrradiationModel


def get_orbital_inclination(time_stamp: float) -> float:
    time_struct = time.gmtime(time_stamp)
    days_in_year = determine_leap_year(time_struct)
    day_number = time_struct.tm_yday
    orbital_inclination = 360 * (day_number / days_in_year)  # in °
    return orbital_inclination


def get_equation_of_time(orbital_inclination) -> float:
    equation_of_time_float = 60 * (0.0066 + 7.3525 * np.cos(np.deg2rad(orbital_inclination + 85.9)) +
                                   9.9359 * np.cos(np.deg2rad(2 * orbital_inclination + 108.9))
                                   + 0.3387 * np.cos(np.deg2rad(3 * orbital_inclination + 105.2)))  # in s
    equation_of_time = int(round(equation_of_time_float))  # in s
    return equation_of_time


def determine_leap_year(time_struct) -> float:
    year = time_struct.tm_year
    if (year % 4) == 0:
        if (year % 100) == 0:
            if (year % 400) == 0:
                days_in_year = 366
            else:
                days_in_year = 365
        else:
            days_in_year = 366
    else:
        days_in_year = 365
    return days_in_year


def get_sun_declination(orbital_inclination) -> float:
    sun_declination = 0.3948 - 23.2559 * np.cos(np.deg2rad(9.1 + orbital_inclination)) - 0.3915 * np.cos(
        np.deg2rad(2 * orbital_inclination + 5.4)) - 0.1764 * np.cos(np.deg2rad(3 * orbital_inclination + 26))  # in °
    return sun_declination


def get_hour_angle(true_local_time) -> float:
    hour_angle = (43200 - true_local_time.tm_hour * 3600 - true_local_time.tm_min * 60 - true_local_time.tm_sec) \
                 * 15 / 3600  # in °
    return hour_angle


def get_extraterrestrial_radiation_horizontal(time_struct, sun_elevation) -> float:
    quantity1 = (1 + 0.03344 * np.cos(np.deg2rad(time_struct.tm_yday * 0.9856 - 2.72))) * 1367  # in [w/m^2]
    extraterrestrial_radiation_horizontal = quantity1 * np.sin(np.deg2rad(sun_elevation))  # in [W/m^2]
    return extraterrestrial_radiation_horizontal


def get_diffuse_radiation_horizontal(global_radiation_horizontal, extraterrestrial_radiation_horizontal,
                                     sun_elevation) -> float:
    # kt is a determining factor
    if sun_elevation > 0:
        kt = global_radiation_horizontal / extraterrestrial_radiation_horizontal
    else:
        kt = 0
    if kt > 0:
        if kt <= 0.3:
            diffuse_radiation_horizontal = global_radiation_horizontal * (
                        1.020 - 0.254 * kt + 0.0123 * np.sin(np.deg2rad(sun_elevation)))  # in [W/m^2]
        elif 1 > kt >= 0.78:
            diffuse_radiation_horizontal = global_radiation_horizontal * (
                        0.486 - 0.182 * np.sin(np.deg2rad(sun_elevation)))  # in [W/m^2]
        elif 0.3 < kt < 0.78:
            diffuse_radiation_horizontal = global_radiation_horizontal * (
                        1.4 - 1.749 * kt + 0.177 * np.sin(np.deg2rad(sun_elevation)))  # in [W/m^2]
        elif kt > 1:
            diffuse_radiation_horizontal = 0
    else:
        diffuse_radiation_horizontal = 0  # in [W/m^2]
    return diffuse_radiation_horizontal


def get_direct_radiation_horizontal(global_radiation_horizontal, diffuse_radiation_horizontal) -> float:
    if diffuse_radiation_horizontal > 0:
        direct_radiation_horizontal = global_radiation_horizontal - diffuse_radiation_horizontal  # in [W/m^2]
    else:
        direct_radiation_horizontal = 0  # in [W/m^2]
    return direct_radiation_horizontal


def get_total_power_radiation(power_radiation_one, power_radiation_two, power_radiation_three,
                              power_radiation_four, power_radiation_five) -> float:
    power_radiation = power_radiation_one + power_radiation_two + power_radiation_three + power_radiation_four + power_radiation_five  # in W
    return power_radiation
    # todo (PL) determine energy for required timestep


class LocationSolarIrradiationModel(SolarIrradiationModel):
    """
        LocationSolarIrradiationModel calculates the total incident solar irradiation on the selected housing object at any
        given time for a specified location.
        Solar irradiation time series data from various sources:
        - German Weather Service (Deutscher Wetterdienst - DWD)
        - Oskar-von-Miller tower - LMU Garching

        """

    def __init__(self, profile_config: ProfileConfig, general_config: GeneralSimulationConfig,
                 housing: Housing):
        super().__init__()
        self.__file: FileProfile = FileProfile(general_config, profile_config.global_horizontal_irradiation_profile_file)
        self.__albedo: float = housing.albedo
        self.__container_azimuth: float = housing.azimuth
        self.__container_outer_layer_absorptivity: float = housing.outer_layer.absorptivity
        self.__small_container_area: float = housing.outer_layer.surface_area_short_side
        self.__tall_container_area: float = housing.outer_layer.surface_area_long_side
        self.__roof_container_area: float = housing.outer_layer.surface_area_roof
        self.__timezone: timezone = self.__file.get_timezone()
        self.__latitude = self.__file.get_latitude()
        self.__longitude = self.__file.get_longitude()

        self.__profile_config: ProfileConfig = profile_config
        self.__general_config: GeneralSimulationConfig = general_config
        self.__housing: Housing = housing

    def get_heat_load(self, time_stamp: float) -> float:
        # time_stamp = float(time_stamp)
        global_radiation_horizontal = self.__file.next(time_stamp)
        orbital_inclination = get_orbital_inclination(time_stamp)
        equation_of_time = get_equation_of_time(orbital_inclination)
        sun_declination = get_sun_declination(orbital_inclination)
        true_local_time = self.__get_true_local_time(equation_of_time, time_stamp)
        hour_angle = get_hour_angle(true_local_time)
        sun_elevation = self.__get_sun_elevation(hour_angle, sun_declination)
        solar_azimuth = self.__get_solar_azimuth(true_local_time, sun_elevation, sun_declination)
        time_struct = time.gmtime(time_stamp)
        extraterrestrial_radiation_horizontal = get_extraterrestrial_radiation_horizontal(time_struct,
                                                                                          sun_elevation)
        diffuse_radiation_horizontal = get_diffuse_radiation_horizontal(global_radiation_horizontal,
                                                                        extraterrestrial_radiation_horizontal,
                                                                        sun_elevation)
        direct_radiation_horizontal = get_direct_radiation_horizontal(global_radiation_horizontal,
                                                                      diffuse_radiation_horizontal)
        reflected_radiation_on_container_sides = self.__get_reflected_radiation_on_container_sides(sun_elevation,
                                                                                                   global_radiation_horizontal)

        # The areas below are defined and identified with the directions they face when azimuth = 0
        power_radiation_container_area_north = self.__get_radiation_container_area_north(
            reflected_radiation_on_container_sides, sun_elevation, solar_azimuth, direct_radiation_horizontal,
            diffuse_radiation_horizontal)
        power_radiation_container_area_south = self.__get_radiation_container_area_south(
            reflected_radiation_on_container_sides, sun_elevation, solar_azimuth, direct_radiation_horizontal,
            diffuse_radiation_horizontal)
        power_radiation_container_area_west = self.__get_radiation_container_area_west(
            reflected_radiation_on_container_sides, sun_elevation, solar_azimuth, direct_radiation_horizontal,
            diffuse_radiation_horizontal)
        power_radiation_container_area_east = self.__get_radiation_container_area_east(
            reflected_radiation_on_container_sides, sun_elevation, solar_azimuth, direct_radiation_horizontal,
            diffuse_radiation_horizontal)
        power_radiation_container_area_roof = self.__get_radiation_container_area_roof(sun_elevation,
                                                                                       direct_radiation_horizontal)
        power_radiation = get_total_power_radiation(power_radiation_container_area_north,
                                                    power_radiation_container_area_south,
                                                    power_radiation_container_area_west,
                                                    power_radiation_container_area_east,
                                                    power_radiation_container_area_roof)
        power_absorbed_radiation = power_radiation * self.__container_outer_layer_absorptivity  # in W
        return power_absorbed_radiation

    def get_global_horizontal_irradiance(self,time_step) -> float:
        return self.__file.next(time_step)  # in W/m2

    def __get_true_local_time(self, equation_of_time, time_stamp):
        local_time: datetime = dt.datetime.fromtimestamp(time_stamp)
        time_zone_data: datetime = self.__timezone.localize(local_time)
        time_zone_offset_seconds: float = time_zone_data.tzinfo._utcoffset.total_seconds()  # in s
        intermediate_quantity = 60 * 4 * (15 - self.__longitude)  # in s
        local_mean_time_epoch = time_stamp - time_zone_offset_seconds + 3600 - intermediate_quantity  # in s
        true_local_time_epoch = local_mean_time_epoch + equation_of_time  # in s
        true_local_time = time.gmtime(true_local_time_epoch)  # as datetime
        return true_local_time

    def __get_sun_elevation(self, hour_angle, sun_declination) -> float:
        sun_elevation = np.rad2deg(np.arcsin(
            np.cos(np.deg2rad(hour_angle)) * np.cos(np.deg2rad(self.__latitude)) * np.cos(
                np.deg2rad(sun_declination)) + np.sin(np.deg2rad(self.__latitude)) * np.sin(
                np.deg2rad(sun_declination))))  # in °
        if sun_elevation < 0:
            sun_elevation = 0  # in °
        return sun_elevation

    def __get_solar_azimuth(self, true_local_time, sun_elevation, sun_declination) -> float:
        quantity1 = np.rad2deg(np.arccos((np.sin(np.deg2rad(sun_elevation)) * np.sin(np.deg2rad(self.__latitude))
                                          - np.sin(np.deg2rad(sun_declination))) / (
                                                     np.cos(np.deg2rad(sun_elevation)) * np.cos(
                                                 np.deg2rad(self.__latitude)))))  # in °
        if sun_elevation >= 0:
            if true_local_time.tm_hour < 12:
                solar_azimuth = 180 - quantity1  # in °
            else:
                solar_azimuth = 180 + quantity1  # in °
        else:
            solar_azimuth = 0  # in °
        return solar_azimuth

    def __get_reflected_radiation_on_container_sides(self, sun_elevation, global_radiation_horizontal) -> float:
        if sun_elevation > 0:
            reflected_radiation_on_container_sides = global_radiation_horizontal * self.__albedo * 0.5  # in [W/m^2]
        else:
            reflected_radiation_on_container_sides = 0  # in [W/m^2]
        return reflected_radiation_on_container_sides

    def __get_radiation_container_area_north(self, reflected_radiation_on_container_sides, sun_elevation,
                                             solar_azimuth, direct_radiation_horizontal,
                                             diffuse_radiation_horizontal) -> float:
        if sun_elevation > 0:
            angle_of_incidence = np.rad2deg(np.arccos(-np.cos(np.deg2rad(sun_elevation)) * np.cos(
                np.deg2rad(solar_azimuth - self.__container_azimuth))))  # in °
            diffuse_radiation = diffuse_radiation_horizontal * 0.5  # in [W/m^2]
        else:
            angle_of_incidence = 0  # in °
            diffuse_radiation = 0  # in [W/m^2]
        if angle_of_incidence > 90:
            angle_of_incidence = 0  # in °
        else:
            angle_of_incidence = angle_of_incidence
        if angle_of_incidence > 0:
            direct_radiation = direct_radiation_horizontal * np.cos(np.deg2rad(angle_of_incidence)) / np.sin(
                np.deg2rad(sun_elevation))  # in [W/m^2]
        else:
            direct_radiation = 0  # in [W/m^2]
        power_radiation_one = self.__small_container_area * (
                    direct_radiation + diffuse_radiation + reflected_radiation_on_container_sides)  # in W
        return power_radiation_one

    def __get_radiation_container_area_south(self, reflected_radiation_on_container_sides, sun_elevation,
                                             solar_azimuth, direct_radiation_horizontal,
                                             diffuse_radiation_horizontal):
        if sun_elevation > 0:
            angle_of_incidence = np.rad2deg(np.arccos(-np.cos(np.deg2rad(sun_elevation)) * np.cos(
                np.deg2rad(solar_azimuth - self.__container_azimuth + 180))))  # in °
            diffuse_radiation = diffuse_radiation_horizontal * 0.5  # in [W/m^2]
        else:
            angle_of_incidence = 0  # in °
            diffuse_radiation = 0  # in [W/m^2]
        if angle_of_incidence > 90:
            angle_of_incidence = 0  # in °
        if angle_of_incidence > 0:
            direct_radiation = direct_radiation_horizontal * np.cos(np.deg2rad(angle_of_incidence)) / np.sin(
                np.deg2rad(sun_elevation))  # in [W/m^2]
        else:
            direct_radiation = 0  # in [W/m^2]
        power_radiation_two = self.__small_container_area * (
                    direct_radiation + diffuse_radiation + reflected_radiation_on_container_sides)  # in W
        return power_radiation_two

    def __get_radiation_container_area_west(self, reflected_radiation_on_container_sides, sun_elevation,
                                            solar_azimuth, direct_radiation_horizontal,
                                            diffuse_radiation_horizontal) -> float:
        if sun_elevation > 0:
            angle_of_incidence = np.rad2deg(np.arccos(-np.cos(np.deg2rad(sun_elevation)) * np.cos(
                np.deg2rad(solar_azimuth - self.__container_azimuth + 90))))  # in °
            diffuse_radiation = diffuse_radiation_horizontal * 0.5  # in [W/m^2]
        else:
            angle_of_incidence = 0  # in °
            diffuse_radiation = 0  # in [W/m^2]
        if angle_of_incidence > 90:
            angle_of_incidence = 0  # in °
        if angle_of_incidence > 0:
            direct_radiation = direct_radiation_horizontal * np.cos(np.deg2rad(angle_of_incidence)) / np.sin(
                np.deg2rad(sun_elevation))  # in [W/m^2]
        else:
            direct_radiation = 0  # in [W/m^2]
        power_radiation_three = self.__tall_container_area * (
                    direct_radiation + diffuse_radiation + reflected_radiation_on_container_sides)  # in W
        return power_radiation_three

    def __get_radiation_container_area_east(self, reflected_radiation_on_container_sides, sun_elevation,
                                            solar_azimuth, direct_radiation_horizontal,
                                            diffuse_radiation_horizontal) -> float:
        if sun_elevation > 0:
            angle_of_incidence = np.rad2deg(np.arccos(-np.cos(np.deg2rad(sun_elevation)) * np.cos(
                np.deg2rad(solar_azimuth - self.__container_azimuth + 270))))  # in °
            diffuse_radiation = diffuse_radiation_horizontal * 0.5  # in [W/m^2]
        else:
            angle_of_incidence = 0  # in °
            diffuse_radiation = 0  # in [W/m^2]
        if angle_of_incidence > 90:
            angle_of_incidence = 0  # in °
        if angle_of_incidence > 0:
            direct_radiation = direct_radiation_horizontal * np.cos(np.deg2rad(angle_of_incidence)) / np.sin(
                np.deg2rad(sun_elevation))  # in [W/m^2]
        else:
            direct_radiation = 0  # in [W/m^2]
        power_radiation_four = self.__tall_container_area * (
                    direct_radiation + diffuse_radiation + reflected_radiation_on_container_sides)  # in W
        return power_radiation_four

    def __get_radiation_container_area_roof(self, sun_elevation, direct_radiation_horizontal) -> float:
        if sun_elevation > 0:
            power_radiation_five = self.__roof_container_area * direct_radiation_horizontal * np.cos(
                np.deg2rad(1 / np.sin(np.deg2rad(sun_elevation))))  # in W
        else:
            power_radiation_five = 0  # in W
        return power_radiation_five

    def create_instance(self) -> SolarIrradiationModel:
        """
        reinstantiates the SolarIrradiationModel
        :return: SolarIrradiationModel
        """
        return LocationSolarIrradiationModel(self.__profile_config, self.__general_config, self.__housing)

    def close(self):
        self.__file.close()




from configparser import ConfigParser
from simses.commons.config.simulation.simulation_config import SimulationConfig, clean_split
from simses.commons.profile.file import FileProfile
from simses.commons.profile.power.file import FilePowerProfile
from simses.commons.utils.utilities import format_float


class ProfileConfig(SimulationConfig):
    """
    Profile specific configs
    """

    SECTION: str = 'PROFILE'

    POWER_PROFILE_DIR: str = 'POWER_PROFILE_DIR'
    TECHNICAL_PROFILE_DIR: str = 'TECHNICAL_PROFILE_DIR'
    THERMAL_PROFILE_DIR: str = 'THERMAL_PROFILE_DIR'
    LOAD_PROFILE: str = 'LOAD_PROFILE'
    FREQUENCY_PROFILE: str = 'FREQUENCY_PROFILE'
    LOAD_FORECAST_PROFILE: str = 'LOAD_FORECAST_PROFILE'
    SOC_PROFILE: str = 'SOC_PROFILE'
    GENERATION_PROFILE: str = 'GENERATION_PROFILE'
    V2G_PROFILE: str = 'V2G_PROFILE'
    BINARY_PROFILE: str = 'BINARY_PROFILE'
    AMBIENT_TEMPERATURE_PROFILE: str = 'AMBIENT_TEMPERATURE_PROFILE'
    BATTERY_TEMPERATURE_PROFILE: str = 'BATTERY_TEMPERATURE_PROFILE'
    GLOBAL_HORIZONTAL_IRRADIATION_PROFILE: str = 'GLOBAL_HORIZONTAL_IRRADIATION_PROFILE'
    LOAD_PROFILE_SCALING: str = 'LOAD_PROFILE_SCALING'
    LOAD_SCALING_FACTOR: str = 'LOAD_SCALING_FACTOR'
    GENERATION_PROFILE_SCALING: str = 'GENERATION_PROFILE_SCALING'
    GENERATION_SCALING_FACTOR: str = 'GENERATION_SCALING_FACTOR'

    def __init__(self, config: ConfigParser, path: str = None):
        super().__init__(path, config)

    @property
    def power_profile_dir(self) -> str:
        """Returns directory of power profiles from __analysis_config file_name"""
        return self.get_data_path(self.get_property(self.SECTION, self.POWER_PROFILE_DIR))

    @property
    def technical_profile_dir(self) -> str:
        """Returns directory of frequency profiles from __analysis_config file_name"""
        return self.get_data_path(self.get_property(self.SECTION, self.TECHNICAL_PROFILE_DIR))

    @property
    def thermal_profile_dir(self) -> str:
        """Returns directory of thermal profiles from __analysis_config file_name"""
        return self.get_data_path(self.get_property(self.SECTION, self.THERMAL_PROFILE_DIR))

    @property
    def load_profile_file(self) -> str:
        """ Return selected load profile file"""
        return self.power_profile_dir + self.get_property(self.SECTION, self.LOAD_PROFILE)

    @property
    def frequency_file(self) -> str:
        """Returns frequency profile file_name name from __analysis_config file_name"""
        return self.technical_profile_dir + self.get_property(self.SECTION, self.FREQUENCY_PROFILE)

    @property
    def load_forecast_file(self) -> str:
        """Returns frequency profile file_name name from __analysis_config file_name"""
        return self.technical_profile_dir + self.get_property(self.SECTION, self.LOAD_FORECAST_PROFILE)

    @property
    def soc_file(self) -> str:
        """Returns soc profile file_name"""
        return self.technical_profile_dir + self.__get_soc_profile()[0]

    @property
    def soc_file_value(self) -> int:
        """Returns soc profile value index"""
        try:
            return int(self.__get_soc_profile()[1])
        except IndexError:
            return 1

    def __get_soc_profile(self) -> [str]:
        return clean_split(self.get_property(self.SECTION, self.SOC_PROFILE), ',')

    @property
    def generation_profile_file(self) -> str:
        """ Return PV generation profile file_name name from __analysis_config file_name"""
        return self.power_profile_dir + self.get_property(self.SECTION, self.GENERATION_PROFILE)

    @property
    def v2g_profile_file(self) -> str:
        """ Return V2G load profile file_name name from __analysis_config file_name"""
        return self.power_profile_dir + self.get_property(self.SECTION, self.V2G_PROFILE)

    @property
    def binary_profile_file(self) -> str:
        """ Return binary profile file_name name from __analysis_config file_name"""
        return self.technical_profile_dir + self.get_property(self.SECTION, self.BINARY_PROFILE)

    @property
    def ambient_temperature_profile_file(self) -> str:
        """ Return selected location ambient temperature profile"""
        return self.thermal_profile_dir + self.get_property(self.SECTION, self.AMBIENT_TEMPERATURE_PROFILE)

    @property
    def battery_temperature_profile_file(self) -> str:
        """ Return selected battery temperature profile"""
        return self.thermal_profile_dir + self.get_property(self.SECTION, self.BATTERY_TEMPERATURE_PROFILE)

    @property
    def global_horizontal_irradiation_profile_file(self) -> str:
        """ Return selected location global horizontal irradiation profile"""
        return self.thermal_profile_dir + self.get_property(self.SECTION, self.GLOBAL_HORIZONTAL_IRRADIATION_PROFILE)

    @property
    def load_profile_scaling(self) -> str:
        """
        Returns the type of load scaling to be applied - Either Energy, Power, or False (if disabled) as str
        """
        return self.get_property(self.SECTION, self.LOAD_PROFILE_SCALING)

    @property
    def load_scaling_factor(self) -> float:
        """ Return scaling factor for the load from __analysis_config file_name"""
        scaling_factor: float = float(self.get_property(self.SECTION, self.LOAD_SCALING_FACTOR))
        if str.lower(self.load_profile_scaling) == 'energy':
            return scaling_factor / self.__get_annual_consumption()
        elif str.lower(self.load_profile_scaling) == 'power':
            return scaling_factor / self.__get_peak_power(self.load_profile_file)
        else:
            return 1.0  # no scaling: original values are taken

    @property
    def generation_profile_scaling(self) -> str:
        """
       Returns the type of generation profile scaling to be applied - Energy, Power, or False (if disabled) as str
       """
        return self.get_property(self.SECTION, self.GENERATION_PROFILE_SCALING)

    @property
    def generation_scaling_factor(self) -> float:
        """ Return scaling factor for pv from __analysis_config file_name"""
        scaling_factor: float = float(self.get_property(self.SECTION, self.GENERATION_SCALING_FACTOR))
        if str.lower(self.generation_profile_scaling) == 'power':
            return scaling_factor / self.__get_peak_power(self.generation_profile_file)
        else:
            return 1.0  # no scaling: original values are taken

    def __get_annual_consumption(self) -> float:
        try:
            value: float = self.__get_value_from_header(FilePowerProfile.Header.ANNUAL_CONSUMPTION, self.load_profile_file)
            return value * 1000.0
        except (IndexError, KeyError):
            return 1.0

    def __get_peak_power(self, filename: str) -> float:
        try:
            peak_power: float = self.__get_value_from_header(FilePowerProfile.Header.PEAK_POWER, filename)
            return peak_power * 1000.0  # Wp
        except (IndexError, KeyError):
            print('Searching for peak power in ' + filename + '... \n(This can take a while. In order to speed things '
                  'up you may include the following header to your profile: # ' + FilePowerProfile.Header.PEAK_POWER +
                  ': [your power value].)')
            value: float = FileProfile.get_max_value_of(filename)
            print('Your power value would be: ' + format_float(value / 1000.0, 6) + ' kWp')
            return value

    @staticmethod
    def __get_value_from_header(name: str, file: str) -> float:
        header: dict = FileProfile.get_header_from(file)
        value = float(header[name])
        return value



import math as m

import numpy as np

from simses.commons.constants import Hydrogen
from simses.technology.hydrogen.hydrogen_storage.hydrogen_storage import HydrogenStorage


class PressureTank(HydrogenStorage):

    __TEMPERATURE = 273.15 + 25  # K
    __WALL_THIKNESS = 1  # cm
    __WALL_DIFFUSION_COEF: float = 4.4 * 10.0 ** (-7) * m.exp(-0.555 / Hydrogen.BOLTZ_CONST / __TEMPERATURE)  # cm2/s  selfdischarge through wall

    def __init__(self, capacity: float, max_pressure: float, soc: float):
        super().__init__()
        self.__soc: float = soc  # p.u.
        self.__capacity: float = capacity  # Wh
        self.__max_pressure: float = max_pressure  # bar
        self.__max_amount_hydrogen: float = self.__get_max_amount_hydrogen()  # mol
        self.__volume: float = self.__get_volume() * (100.0 * 100.0 * 100.0)  # cm3
        self.__inner_radius: float = (3.0 * self.__volume / (4.0 * m.pi)) ** (1.0 / 3.0)  # cm
        self.__inner_surface: float = 4.0 * m.pi * self.__inner_radius ** 2  # cm2
        self.__tank_pressure: float = 0.0
        self.__calculate_tank_pressure()

    def __get_max_amount_hydrogen(self) -> float:
        """ returns max amount of hydrogen in mol that can be stored within the given pressuretank"""
        return self.__capacity / 1000.0 / (Hydrogen.LOWER_HEATING_VALUE * 2 * Hydrogen.MOLAR_MASS_HYDROGEN)

    def __get_volume(self) -> float:
        """ returns required volume in m³ of the tank that represents the desired capacity """
        pressure = self.__max_pressure * 10**5
        a = 1.0
        b = - (self.__max_amount_hydrogen * Hydrogen.VAN_D_WAALS_COEF_B + self.__max_amount_hydrogen * Hydrogen.IDEAL_GAS_CONST \
               * self.__TEMPERATURE / pressure)
        c = Hydrogen.VAN_D_WAALS_COEF_A * self.__max_amount_hydrogen ** 2 / pressure
        d = - Hydrogen.VAN_D_WAALS_COEF_B * Hydrogen.VAN_D_WAALS_COEF_A * self.__max_amount_hydrogen ** 3 / pressure
        coeff = [a, b, c, d]
        n = np.roots(coeff)
        return np.max(np.real(n))

    def __get_hydrogen_amount(self) -> float:
        """ returns the current amount of hydrogen within the tank for a given SOC """
        return self.__max_amount_hydrogen * self.__soc

    def __get_hydrogen_concentration(self) -> float:
        """ returns the current hydrogen concentration within the tank in mol/cm3 """
        return self.__get_hydrogen_amount() / self.__volume

    def __ideal_gas_law(self, act_mass, volume) -> float:
        """ returns current pressure of the tank in Pa
            volume in m3, temperature in K, act_mass in mol """
        return act_mass * Hydrogen.IDEAL_GAS_CONST * self.__TEMPERATURE / \
               (volume - act_mass * Hydrogen.VAN_D_WAALS_COEF_B) - Hydrogen.VAN_D_WAALS_COEF_A * \
               act_mass ** 2 / volume ** 2  # pressure in Pa

    def __diff_loss_wall(self) -> float:
        """ returns mass losses caused by diffusion through the wall in mol/s """
        return self.__WALL_DIFFUSION_COEF * self.__get_hydrogen_concentration() * self.__inner_surface / self.__WALL_THIKNESS

    def calculate_soc(self, time_difference, hydrogen_net_flow: float) -> None:
        self.__soc += (hydrogen_net_flow - self.__diff_loss_wall()) * (time_difference) / self.__max_amount_hydrogen
        self.__calculate_tank_pressure()

    def __calculate_tank_pressure(self) -> None:
        act_mass = self.__get_hydrogen_amount()
        volume = self.__volume / (100 * 100 * 100)  # cm³ -> m³
        self.__tank_pressure = self.__ideal_gas_law(act_mass, volume) / 10 ** 5  # pressure in bar

    def get_soc(self):
        return self.__soc

    def get_capacity(self):
        return self.__capacity

    def get_tank_pressure(self) -> float:
        return self.__tank_pressure

    def close(self) -> None:
        pass



import os
from configparser import ConfigParser
from datetime import datetime, timezone
import pytest
from simses.commons.config.simulation.general import GeneralSimulationConfig
from simses.commons.profile.power.file import FilePowerProfile
from simses.commons.utils.utilities import remove_file

DELIMITER: str = ','
HEADER: str = '# Unit: W'
START: int = 0
END: int = 60
STEP: int = 1


def create_file(value):
    filename: str = os.path.join(os.path.dirname(__file__), 'mockup_power_profile.csv')
    with open(filename, mode='w') as file:
        file.write(HEADER + '\n')
        # add bad inputs
        file.write('0;0')
        file.write('\n')
        file.write('#\n')
        file.write('"""\n')
        file.write('"\n')
        file.write('   #\n')
        # add good inputs
        tstmp = START
        while tstmp <= END:
            file.write(str(tstmp) + DELIMITER + str(value) + '\n')
            tstmp += STEP
    return filename


def create_config():
    conf: ConfigParser = ConfigParser()
    conf.add_section('GENERAL')
    date = datetime.fromtimestamp(START, timezone.utc)
    conf.set('GENERAL', 'START', date.strftime("%Y-%m-%d %H:%M:%S"))
    date = datetime.fromtimestamp(END, timezone.utc)
    conf.set('GENERAL', 'END', date.strftime("%Y-%m-%d %H:%M:%S"))
    conf.set('GENERAL', 'TIME_STEP', str(STEP))
    return GeneralSimulationConfig(config=conf)


@pytest.fixture(scope='function')
def uut(power, scaling_factor):
    filename = create_file(power)
    uut = FilePowerProfile(config=create_config(), filename=filename, delimiter=DELIMITER, scaling_factor=scaling_factor)
    yield uut
    uut.close()
    remove_file(filename)


@pytest.mark.parametrize('power, time_factor, scaling_factor, result',
                         [
                            (0, 1, 0, 0),
                            (0, 1, 2, 0),
                            (1, 1, 1, 1),
                            (2, 1, 3, 6),
                            (6, 0.5, 0.5, 3),
                            (1, 5, 1, 1),
                            (0, 1, 1, 0)
                         ]
                         )
def test_next(time_factor, result, uut):
    tstmp = START
    while tstmp < END:
        assert abs(uut.next(tstmp) - result) <= 1e-10
        tstmp += STEP * time_factor



from math import log10

from simses.commons.config.simulation.redox_flow import RedoxFlowConfig
from simses.commons.log import Logger
from simses.commons.state.technology.redox_flow import RedoxFlowState
from simses.technology.redox_flow.stack.electrolyte.abstract_electrolyte import ElectrolyteSystem
from simses.technology.redox_flow.stack.abstract_stack import StackModule


class DummyStack3000W(StackModule):
    """DummyStack3000W describes a dummy stack with a constant internal resistance. The dependency of the
    resistance from SOC, flow rate and current is neglected. This stack type can be used to easily implement other stack
    parameters."""

    """The nominal power is calculated for an SOC range from 20 % to 80 % and the listed parameters. It corresponds to 
    the power that can be obtained over this SOC range."""
    __STACK_POWER_NOM = 3000  # W
    __CELL_NUMBER = 20  # -
    __INTERNAL_RESISTANCE_CELL = 1  # Ohm cm^2
    __CELL_AREA = 1000  # cm^2
    __ELECTRODE_THICKNESS = 0.37  # cm
    __ELECTRODE_POROSITY = 0.9  # -
    __MIN_CELL_VOLTAGE = 1.2  # V
    __MAX_CELL_VOLTAGE = 1.6  # V
    __SELF_DISCHARGE_CURRENT_DENS = 1  # mA/cm^2
    __HYDRAULIC_RESISTANCE = 5.969E10  # 1/m^3
    __DEPENDENT_PARAMETERS = False

    def __init__(self, electrolyte_system: ElectrolyteSystem, voltage: float, power: float,
                 redox_flow_config: RedoxFlowConfig):
        super().__init__(electrolyte_system, voltage, power, self.__CELL_NUMBER, self.__STACK_POWER_NOM,
                         redox_flow_config)
        self.__log: Logger = Logger(__name__)
        self.__electrolyte_system: ElectrolyteSystem = electrolyte_system

    def get_open_circuit_voltage(self, redox_flow_state: RedoxFlowState) -> float:
        """
        Literature source: Fink, Holger. Untersuchung von Verlustmechanismen in Vanadium-Flussbatterien. Diss.
        Technische Universität München, 2019.
        equation 5.18, assumption: SOH = 100 %, therefore ver = 0.5
        """
        concentration_v = self.__electrolyte_system.get_vanadium_concentration()
        soc_stack = redox_flow_state.soc_stack
        # equation is defined for SOC = ]0-1[
        if soc_stack > 0.999:
            soc_stack = 0.999
        if soc_stack < 0.001:
            soc_stack = 0.001
        temperature = redox_flow_state.temperature
        concentration_h_start = 2.6  # mol/l
        ocv_cell = (1.255 + 0.07 + 0.059 * temperature / 298.15 * log10((soc_stack / (1 - soc_stack) *
                    (concentration_h_start + concentration_v / 1000 * (soc_stack + 0.5)))**2 * (concentration_h_start +
                    concentration_v / 1000 * (soc_stack - 0.5))))
        self.__log.debug('OCV cell: ' + str(ocv_cell) + ' V')
        return ocv_cell * self.get_cell_per_stack() * self.get_serial_scale()

    def get_nominal_voltage_cell(self) -> float:
        """Calculated for a temperature of 25 °C and at SOC 50 %."""
        nominal_voltage_cell = 1.423
        return nominal_voltage_cell

    def get_internal_resistance(self, redox_flow_state: RedoxFlowState) -> float:
        resistance = self.__INTERNAL_RESISTANCE_CELL / self.get_specific_cell_area()
        return resistance * self.get_cell_per_stack() * self.get_serial_scale() / self.get_parallel_scale()

    def get_cell_per_stack(self) -> int:
        return self.__CELL_NUMBER

    def get_min_voltage(self) -> float:
        return self.__MIN_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_max_voltage(self) -> float:
        return self.__MAX_CELL_VOLTAGE * self.get_cell_per_stack() * self.get_serial_scale()

    def get_self_discharge_current(self, redox_flow_state: RedoxFlowState) -> float:
        return (self.__SELF_DISCHARGE_CURRENT_DENS / 1000 * self.get_specific_cell_area() * self.get_cell_per_stack() *
                self.get_serial_scale() * self.get_parallel_scale())

    def get_specific_cell_area(self) -> float:
        return self.__CELL_AREA

    def get_electrode_thickness(self) -> float:
        return self.__ELECTRODE_THICKNESS / 100

    def get_electrode_porosity(self) -> float:
        return self.__ELECTRODE_POROSITY

    def get_hydraulic_resistance(self) -> float:
        return self.__HYDRAULIC_RESISTANCE / (self.get_serial_scale() * self.get_parallel_scale())

    def dependent_parameters(self) -> bool:
        return self.__DEPENDENT_PARAMETERS

    def close(self):
        super().close()
        self.__log.close()
        self.__electrolyte_system.close()
